import argparse
import logging
import pathlib
from typing import List, Optional

import rasa.cli.utils
import rasa.shared.core.domain
import rasa.shared.data
import rasa.shared.nlu.training_data.loading
import rasa.shared.nlu.training_data.util
import rasa.shared.utils.cli
import rasa.shared.utils.io
import rasa.utils.common
from rasa import telemetry
from rasa.cli import SubParsersAction
from rasa.cli.arguments import data as arguments
from rasa.cli.arguments import default_arguments
from rasa.cli.validation.bot_config import validate_files
from rasa.cli.validation.config_path_validation import get_validated_path
from rasa.core.config.configuration import Configuration
from rasa.e2e_test.e2e_config import create_llm_e2e_test_converter_config
from rasa.e2e_test.e2e_test_converter import E2ETestConverter
from rasa.e2e_test.utils.e2e_yaml_utils import E2ETestYAMLWriter
from rasa.shared.constants import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_DATA_PATH,
    DEFAULT_DOMAIN_PATHS,
)
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.common import minimal_kwargs
from rasa.shared.utils.yaml import read_yaml_file, write_yaml
from rasa.utils.beta import ensure_beta_feature_is_enabled

logger = logging.getLogger(__name__)

RASA_PRO_BETA_E2E_CONVERSION_ENV_VAR_NAME = "RASA_PRO_BETA_E2E_CONVERSION"


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all data parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    data_parser = subparsers.add_parser(
        "data",
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Utils for the Rasa training files.",
    )
    data_parser.set_defaults(func=lambda _: data_parser.print_help(None))

    data_subparsers = data_parser.add_subparsers()

    _add_data_convert_parsers(data_subparsers, parents)
    _add_data_split_parsers(data_subparsers, parents)
    _add_data_validate_parsers(data_subparsers, parents)
    _add_data_migrate_parsers(data_subparsers, parents)


def _add_data_convert_parsers(
    data_subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    convert_parser = data_subparsers.add_parser(
        "convert",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Converts Rasa data between different formats.",
    )
    convert_parser.set_defaults(func=lambda _: convert_parser.print_help(None))

    convert_subparsers = convert_parser.add_subparsers()
    convert_nlu_parser = convert_subparsers.add_parser(
        "nlu",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Converts NLU data between formats.",
    )
    convert_nlu_parser.set_defaults(func=_convert_nlu_data)
    arguments.set_convert_nlu_arguments(convert_nlu_parser, data_type="Rasa NLU")

    convert_e2e_parser = convert_subparsers.add_parser(
        "e2e",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Convert input sample conversations into E2E test cases.",
    )
    convert_e2e_parser.set_defaults(func=convert_data_to_e2e_tests)
    arguments.set_convert_e2e_arguments(convert_e2e_parser)


def _add_data_split_parsers(
    data_subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    split_parser = data_subparsers.add_parser(
        "split",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Splits Rasa data into training and test data.",
    )
    split_parser.set_defaults(func=lambda _: split_parser.print_help(None))

    split_subparsers = split_parser.add_subparsers()
    nlu_split_parser = split_subparsers.add_parser(
        "nlu",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Performs a split of your NLU data into training and test data "
        "according to the specified percentages.",
    )
    nlu_split_parser.set_defaults(func=split_nlu_data)

    arguments.set_split_arguments(nlu_split_parser)

    stories_split_parser = split_subparsers.add_parser(
        "stories",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Performs a split of your stories into training and test data "
        "according to the specified percentages.",
    )
    stories_split_parser.set_defaults(func=split_stories_data)

    arguments.set_split_arguments(stories_split_parser)


def _add_data_validate_parsers(
    data_subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    validate_parser = data_subparsers.add_parser(
        "validate",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Validates domain and data files to check for possible mistakes.",
    )
    _append_story_structure_arguments(validate_parser)
    validate_parser.set_defaults(
        func=lambda args: _validate_files_with_configuration(
            args.fail_on_warnings,
            args.max_history,
            _build_training_data_importer(args),
            sub_agents=args.sub_agents,
            endpoints=args.endpoints,
        )
    )
    arguments.set_validator_arguments(validate_parser)

    validate_subparsers = validate_parser.add_subparsers()
    story_structure_parser = validate_subparsers.add_parser(
        "stories",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Checks for inconsistencies in the story files.",
    )
    _append_story_structure_arguments(story_structure_parser)

    story_structure_parser.set_defaults(
        func=lambda args: _validate_files_with_configuration(
            args.fail_on_warnings,
            args.max_history,
            _build_training_data_importer(args),
            sub_agents=args.sub_agents,
            endpoints=args.endpoints,
            stories_only=True,
        )
    )
    arguments.set_validator_arguments(story_structure_parser)

    flows_structure_parser = validate_subparsers.add_parser(
        "flows",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Checks for inconsistencies in the flows files.",
    )
    flows_structure_parser.set_defaults(
        func=lambda args: _validate_files_with_configuration(
            args.fail_on_warnings,
            args.max_history,
            _build_training_data_importer(args),
            sub_agents=args.sub_agents,
            endpoints=args.endpoints,
            flows_only=True,
        )
    )
    arguments.set_validator_arguments(flows_structure_parser)

    translations_structure_parser = validate_subparsers.add_parser(
        "translations",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Checks for inconsistencies of the flow and response translation.",
    )
    translations_structure_parser.set_defaults(
        func=lambda args: _validate_files_with_configuration(
            args.fail_on_warnings,
            args.max_history,
            _build_training_data_importer(args),
            sub_agents=args.sub_agents,
            endpoints=args.endpoints,
            translations_only=True,
        )
    )
    arguments.set_validator_arguments(translations_structure_parser)


def _build_training_data_importer(args: argparse.Namespace) -> "TrainingDataImporter":
    config = get_validated_path(
        args.config, "config", DEFAULT_CONFIG_PATH, none_is_valid=True
    )

    # Exit the validation if the domain path is invalid
    domain = get_validated_path(
        args.domain, "domain", DEFAULT_DOMAIN_PATHS, none_is_valid=False
    )

    if config:
        return TrainingDataImporter.load_from_config(
            domain_path=domain, training_data_paths=args.data, config_path=config
        )
    else:
        return TrainingDataImporter.load_from_dict(
            domain_path=domain, training_data_paths=args.data, config_path=config
        )


def _append_story_structure_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--max-history",
        type=int,
        default=None,
        help="Number of turns taken into account for story structure validation.",
    )
    default_arguments.add_config_param(parser)


def split_nlu_data(args: argparse.Namespace) -> None:
    """Load data from a file path and split the NLU data into test and train examples.

    Args:
        args: Commandline arguments
    """
    data_path = get_validated_path(args.nlu, "nlu", DEFAULT_DATA_PATH)
    data_path = rasa.shared.data.get_nlu_directory(data_path)

    nlu_data = rasa.shared.nlu.training_data.loading.load_data(data_path)
    extension = rasa.shared.nlu.training_data.util.get_file_format_extension(data_path)

    train, test = nlu_data.train_test_split(args.training_fraction, args.random_seed)

    train.persist(args.out, filename=f"training_data{extension}")
    test.persist(args.out, filename=f"test_data{extension}")

    telemetry.track_data_split(args.training_fraction, "nlu")


def split_stories_data(args: argparse.Namespace) -> None:
    """Load data from a file path and split stories into test and train examples.

    Args:
        args: Commandline arguments
    """
    from sklearn.model_selection import train_test_split

    from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
        KEY_STORIES,
        YAMLStoryReader,
    )

    data_path = get_validated_path(args.nlu, "nlu", DEFAULT_DATA_PATH)
    data_files = rasa.shared.data.get_data_files(
        data_path, YAMLStoryReader.is_stories_file
    )
    out_path = pathlib.Path(args.out)
    out_path.mkdir(parents=True, exist_ok=True)

    # load Yaml stories data
    for file_name in data_files:
        file_data = read_yaml_file(file_name)
        assert isinstance(file_data, dict)
        stories = file_data.get(KEY_STORIES, [])
        if not stories:
            logger.info(f"File {file_name} has no stories, skipped")
            continue

        file_path = pathlib.Path(file_name)

        # everything besides stories are going into the training data
        train, test = train_test_split(
            stories, test_size=1 - args.training_fraction, random_state=args.random_seed
        )
        out_file_train = out_path / ("train_" + file_path.name)
        out_file_test = out_path / ("test_" + file_path.name)

        # train file contains everything else from the file + train stories
        file_data[KEY_STORIES] = train
        write_yaml(file_data, out_file_train)

        # test file contains just test stories
        write_yaml({KEY_STORIES: test}, out_file_test)
        logger.info(
            f"From {file_name} we produced {out_file_train} "
            f"with {len(train)} stories and {out_file_test} "
            f"with {len(test)} stories"
        )


def _convert_nlu_data(args: argparse.Namespace) -> None:
    import rasa.nlu.convert

    if args.format in ["json", "yaml"]:
        rasa.nlu.convert.convert_training_data(
            args.data, args.out, args.format, args.language
        )
        telemetry.track_data_convert(args.format, "nlu")
    else:
        rasa.shared.utils.cli.print_error_and_exit(
            "Could not recognize output format. Supported output formats: 'json' "
            "and 'yaml'. Specify the desired output format with '--format'."
        )


def _add_data_migrate_parsers(
    data_subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    migrate_parser = data_subparsers.add_parser(
        "migrate",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=parents,
        help="Converts Rasa domain 2.0 format to required format for 3.0.",
    )
    migrate_parser.set_defaults(func=_migrate_domain)

    arguments.set_migrate_arguments(migrate_parser)


def _migrate_domain(args: argparse.Namespace) -> None:
    import rasa.core.migrate

    rasa.core.migrate.migrate_domain_format(args.domain, args.out)


def validate_e2e_test_conversion_output_path(output_path: str) -> None:
    """Validates that the provided output path is within the project directory.

    Args:
        output_path (str): The output path to be validated.

    Raises:
        RasaException: If the provided output path is an absolute path.
    """
    if pathlib.Path(output_path).is_absolute():
        raise RasaException(
            "Please provide a relative output path within the assistant "
            "project directory in which the command is running."
        )


def convert_data_to_e2e_tests(args: argparse.Namespace) -> None:
    """Converts sample conversation data into E2E test cases
    and stores them in the output YAML file.

    Args:
        args: The arguments passed in from the CLI.
    """
    try:
        ensure_beta_feature_is_enabled(
            "conversion of sample conversations into end-to-end tests",
            RASA_PRO_BETA_E2E_CONVERSION_ENV_VAR_NAME,
        )
        validate_e2e_test_conversion_output_path(args.output)

        config_path = pathlib.Path(args.output)
        llm_config = create_llm_e2e_test_converter_config(config_path)

        kwargs = minimal_kwargs(vars(args), E2ETestConverter)
        converter = E2ETestConverter(llm_config=llm_config, **kwargs)
        yaml_tests_string = converter.run()

        writer = E2ETestYAMLWriter(output_path=args.output)
        writer.write_to_file(yaml_tests_string)
    except RasaException as exc:
        rasa.shared.utils.cli.print_error_and_exit(
            f"Failed to convert the data into E2E tests. Error: {exc}"
        )


def _initialize_configuration_for_validation(
    sub_agents: Optional[str] = None, endpoints: Optional[str] = None
) -> None:
    """Initialize Configuration before validation.

    Args:
        sub_agents: Path to sub-agents directory for validation.
        endpoints: Path to the endpoints configuration file.
    """
    if endpoints:
        Configuration.initialise_endpoints(endpoints_path=pathlib.Path(endpoints))
    if sub_agents:
        Configuration.initialise_sub_agents(pathlib.Path(sub_agents))
    if not endpoints and not sub_agents:
        Configuration.initialise_empty()


def _validate_files_with_configuration(
    fail_on_warnings: bool,
    max_history: int,
    importer: TrainingDataImporter,
    sub_agents: Optional[str] = None,
    endpoints: Optional[str] = None,
    stories_only: bool = False,
    flows_only: bool = False,
    translations_only: bool = False,
) -> None:
    """Wrapper for validate_files that ensures Configuration is initialized first.

    Args:
        fail_on_warnings: `True` if the process should exit with a non-zero status
        max_history: The max history to use when validating the story structure.
        importer: The `TrainingDataImporter` to use to load the training data.
        sub_agents: Path to sub-agents directory for validation.
        endpoints: Path to the endpoints configuration file.
        stories_only: If `True`, only the story structure is validated.
        flows_only: If `True`, only the flows are validated.
        translations_only: If `True`, only the translations data is validated.
    """
    # Initialize Configuration before calling validate_files
    _initialize_configuration_for_validation(sub_agents=sub_agents, endpoints=endpoints)

    # Call the original validate_files function
    validate_files(
        fail_on_warnings=fail_on_warnings,
        max_history=max_history,
        importer=importer,
        stories_only=stories_only,
        flows_only=flows_only,
        translations_only=translations_only,
        sub_agents=sub_agents,
        endpoints=endpoints,
    )



from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Union

import numpy as np
import tensorflow as tf

from rasa.nlu.constants import DEFAULT_TRANSFORMER_SIZE
from rasa.shared.nlu.constants import TEXT
from rasa.utils.tensorflow import layers
from rasa.utils.tensorflow.constants import (
    CONCAT_DIMENSION,
    CONNECTION_DENSITY,
    DENSE_DIMENSION,
    DENSE_INPUT_DROPOUT,
    DROP_RATE,
    DROP_RATE_ATTENTION,
    HIDDEN_LAYERS_SIZES,
    KEY_RELATIVE_ATTENTION,
    MASKED_LM,
    MAX_RELATIVE_POSITION,
    NUM_HEADS,
    NUM_TRANSFORMER_LAYERS,
    REGULARIZATION_CONSTANT,
    SENTENCE,
    SEQUENCE,
    SPARSE_INPUT_DROPOUT,
    TRANSFORMER_SIZE,
    UNIDIRECTIONAL_ENCODER,
    VALUE_RELATIVE_ATTENTION,
)
from rasa.utils.tensorflow.exceptions import TFLayerConfigException
from rasa.utils.tensorflow.model_data import FeatureSignature
from rasa.utils.tensorflow.transformer import TransformerEncoder


class RasaCustomLayer(tf.keras.layers.Layer):
    """Parent class for all classes in `rasa_layers.py`.

    Allows a shared implementation for adjusting `DenseForSparse`
    layers during incremental training.

    During fine-tuning, sparse feature sizes might change due to addition of new data.
    If this happens, we need to adjust our `DenseForSparse` layers to a new size.
    `ConcatenateSparseDenseFeatures`, `RasaSequenceLayer` and
    `RasaFeatureCombiningLayer` all inherit from `RasaCustomLayer` and thus can
    change their own `DenseForSparse` layers if it's needed.
    """

    def adjust_sparse_layers_for_incremental_training(
        self,
        new_sparse_feature_sizes: Dict[Text, Dict[Text, List[int]]],
        old_sparse_feature_sizes: Dict[Text, Dict[Text, List[int]]],
        reg_lambda: float,
    ) -> None:
        """Finds and adjusts `DenseForSparse` layers during incremental training.

        Recursively looks through the layers until it finds all the `DenseForSparse`
        ones and adjusts those which have their sparse feature sizes increased.

        This function heavily relies on the name of `DenseForSparse` layer being
        in the following format - f"sparse_to_dense.{attribute}_{feature_type}" -
        in order to correctly extract the attribute and feature type.

        New and old sparse feature sizes could look like this:
        {TEXT: {FEATURE_TYPE_SEQUENCE: [4, 24, 128], FEATURE_TYPE_SENTENCE: [4, 128]}}

        Args:
            new_sparse_feature_sizes: sizes of current sparse features.
            old_sparse_feature_sizes: sizes of sparse features the model was
                                      previously trained on.
            reg_lambda: regularization constant.
        """
        for name, layer in self._tf_layers.items():
            if isinstance(layer, RasaCustomLayer):
                layer.adjust_sparse_layers_for_incremental_training(
                    new_sparse_feature_sizes=new_sparse_feature_sizes,
                    old_sparse_feature_sizes=old_sparse_feature_sizes,
                    reg_lambda=reg_lambda,
                )
            elif isinstance(layer, layers.DenseForSparse):
                attribute = layer.get_attribute()
                feature_type = layer.get_feature_type()
                if (
                    attribute in new_sparse_feature_sizes
                    and feature_type in new_sparse_feature_sizes[attribute]
                ):
                    new_feature_sizes = new_sparse_feature_sizes[attribute][
                        feature_type
                    ]
                    old_feature_sizes = old_sparse_feature_sizes[attribute][
                        feature_type
                    ]
                    if sum(new_feature_sizes) > sum(old_feature_sizes):
                        self._tf_layers[name] = self._replace_dense_for_sparse_layer(
                            layer_to_replace=layer,
                            new_sparse_feature_sizes=new_feature_sizes,
                            old_sparse_feature_sizes=old_feature_sizes,
                            attribute=attribute,
                            feature_type=feature_type,
                            reg_lambda=reg_lambda,
                        )

    @staticmethod
    def _replace_dense_for_sparse_layer(
        layer_to_replace: layers.DenseForSparse,
        new_sparse_feature_sizes: List[int],
        old_sparse_feature_sizes: List[int],
        attribute: Text,
        feature_type: Text,
        reg_lambda: float,
    ) -> layers.DenseForSparse:
        """Replaces a `DenseForSparse` layer with a new one.

        Replaces an existing `DenseForSparse` layer with a new one
        in order to adapt it to incremental training.

        Args:
            layer_to_replace: a `DenseForSparse` layer that is used to create a new one.
            new_sparse_feature_sizes: sizes of sparse features that will be
                                      the input of the layer.
            old_sparse_feature_sizes: sizes of sparse features that used to be
                                      the input of the layer.
            attribute: an attribute of the data fed to the layer.
            feature_type: a feature type of the data fed to the layer.
            reg_lambda: regularization constant.

        Returns:
            New `DenseForSparse` layer.
        """
        kernel = layer_to_replace.get_kernel().numpy()
        bias = layer_to_replace.get_bias()
        if bias is not None:
            bias = bias.numpy()
        units = layer_to_replace.get_units()
        # split kernel by feature sizes to update the layer accordingly
        kernel_splits = []
        splitting_index = 0
        for size in old_sparse_feature_sizes:
            kernel_splits.append(kernel[splitting_index : splitting_index + size, :])
            splitting_index += size
        additional_sizes = [
            new_size - old_size
            for new_size, old_size in zip(
                new_sparse_feature_sizes, old_sparse_feature_sizes
            )
        ]
        std, mean = np.std(kernel), np.mean(kernel)
        additional_weights = [
            np.random.normal(mean, std, size=(num_rows, units)).astype(np.float32)
            for num_rows in additional_sizes
        ]
        merged_weights = [
            np.vstack((existing, new))
            for existing, new in zip(kernel_splits, additional_weights)
        ]
        # stack each merged weight to form a new weight tensor
        new_weights = np.vstack(merged_weights)
        kernel_init = tf.constant_initializer(new_weights)
        bias_init = tf.constant_initializer(bias) if bias is not None else None
        new_layer = layers.DenseForSparse(
            name=f"sparse_to_dense.{attribute}_{feature_type}",
            reg_lambda=reg_lambda,
            units=units,
            use_bias=bias is not None,
            kernel_initializer=kernel_init,
            bias_initializer=bias_init,
        )
        return new_layer


class ConcatenateSparseDenseFeatures(RasaCustomLayer):
    """Combines multiple sparse and dense feature tensors into one dense tensor.

    This layer combines features from various featurisers into a single feature array
    per input example. All features must be of the same feature type, i.e. sentence-
    level or sequence-level (token-level).

    The layer combines a given list of tensors (whether sparse or dense) by:
    1. converting sparse tensors into dense ones
    2. optionally, applying dropout to sparse tensors before and/or after the conversion
    3. concatenating all tensors along the last dimension

    Arguments:
        attribute: Name of attribute (e.g. `text` or `label`) whose features will be
            processed.
        feature_type: Feature type to be processed -- `sequence` or `sentence`.
        feature_type_signature: A list of signatures for the given attribute and feature
            type.
        config: A model config for correctly parametrising the layer.

    Input shape:
        Tuple containing one list of N-D tensors, each with shape: `(batch_size, ...,
        input_dim)`.
        All dense tensors must have the same shape, except possibly the last dimension.
        All sparse tensors must have the same shape, including the last dimension.

    Output shape:
        N-D tensor with shape: `(batch_size, ..., units)` where `units` is the sum of
        the last dimension sizes across all input tensors, with sparse tensors instead
        contributing `config[DENSE_DIMENSION][attribute]` units each.

    Raises:
        A `TFLayerConfigException` if no feature signatures are provided.

    Attributes:
        output_units: The last dimension size of the layer's output.
    """

    SPARSE_DROPOUT = "sparse_dropout"
    SPARSE_TO_DENSE = "sparse_to_dense"
    DENSE_DROPOUT = "dense_dropout"

    def __init__(
        self,
        attribute: Text,
        feature_type: Text,
        feature_type_signature: List[FeatureSignature],
        config: Dict[Text, Any],
    ) -> None:
        """Creates a new `ConcatenateSparseDenseFeatures` object."""
        if not feature_type_signature:
            raise TFLayerConfigException(
                "The feature type signature must contain some feature signatures."
            )

        super().__init__(
            name=f"concatenate_sparse_dense_features_{attribute}_{feature_type}"
        )

        self._check_sparse_input_units(feature_type_signature)

        self.output_units = self._calculate_output_units(
            attribute, feature_type_signature, config
        )

        # Prepare dropout and sparse-to-dense layers if any sparse tensors are expected
        self._tf_layers: Dict[Text, tf.keras.layers.Layer] = {}
        if any([signature.is_sparse for signature in feature_type_signature]):
            self._prepare_layers_for_sparse_tensors(attribute, feature_type, config)

    def _check_sparse_input_units(
        self, feature_type_signature: List[FeatureSignature]
    ) -> None:
        """Checks that all sparse features have the same last dimension size."""
        sparse_units = [
            feature_sig.units
            for feature_sig in feature_type_signature
            if feature_sig.is_sparse
        ]
        if len(set(sparse_units)) > 1:
            raise TFLayerConfigException(
                f"All sparse features must have the same last dimension size but found "
                f"different sizes: {set(sparse_units)}."
            )

    def _prepare_layers_for_sparse_tensors(
        self, attribute: Text, feature_type: Text, config: Dict[Text, Any]
    ) -> None:
        """Sets up sparse tensor pre-processing before combining with dense ones."""
        # For optionally applying dropout to sparse tensors
        if config[SPARSE_INPUT_DROPOUT]:
            self._tf_layers[self.SPARSE_DROPOUT] = layers.SparseDropout(
                rate=config[DROP_RATE]
            )

        # For converting sparse tensors to dense
        self._tf_layers[self.SPARSE_TO_DENSE] = layers.DenseForSparse(
            name=f"sparse_to_dense.{attribute}_{feature_type}",
            units=config[DENSE_DIMENSION][attribute],
            reg_lambda=config[REGULARIZATION_CONSTANT],
        )

        # For optionally apply dropout to sparse tensors after they're converted to
        # dense tensors.
        if config[DENSE_INPUT_DROPOUT]:
            self._tf_layers[self.DENSE_DROPOUT] = tf.keras.layers.Dropout(
                rate=config[DROP_RATE]
            )

    @staticmethod
    def _calculate_output_units(
        attribute: Text,
        feature_type_signature: List[FeatureSignature],
        config: Dict[Text, Any],
    ) -> int:
        """Determines the output units from the provided feature signatures.

        Sparse features will be turned into dense ones, hence they each contribute with
        their future dense number of units.
        """
        return sum(
            [
                config[DENSE_DIMENSION][attribute]
                if signature.is_sparse
                else (signature.units or 0)
                for signature in feature_type_signature
            ]
        )

    def _process_sparse_feature(
        self, feature: tf.SparseTensor, training: bool
    ) -> tf.Tensor:
        """Turns sparse tensor into dense, possibly adds dropout before and/or after."""
        if self.SPARSE_DROPOUT in self._tf_layers:
            feature = self._tf_layers[self.SPARSE_DROPOUT](feature, training=training)

        feature = self._tf_layers[self.SPARSE_TO_DENSE](feature)

        if self.DENSE_DROPOUT in self._tf_layers:
            feature = self._tf_layers[self.DENSE_DROPOUT](feature, training=training)

        return feature

    def call(
        self,
        inputs: Tuple[List[Union[tf.Tensor, tf.SparseTensor]]],
        training: bool = False,
    ) -> tf.Tensor:
        """Combines sparse and dense feature tensors into one tensor.

        Arguments:
            inputs: Contains the input tensors, all of the same rank.
            training: A flag indicating whether the layer should behave in training mode
                (applying dropout to sparse tensors if applicable) or in inference mode
                (not applying dropout).

        Returns:
            Single tensor with all input tensors combined along the last dimension.
        """
        features = inputs[0]

        dense_features = []
        for f in features:
            if isinstance(f, tf.SparseTensor):
                f = self._process_sparse_feature(f, training)
            dense_features.append(f)

        # Now that all features are made dense, concatenate them along the last (units)
        # dimension.
        return tf.concat(dense_features, axis=-1)


class RasaFeatureCombiningLayer(RasaCustomLayer):
    """Combines multiple dense or sparse feature tensors into one.

    This layer combines features by following these steps:
    1. Apply a `ConcatenateSparseDenseFeatures` layer separately to sequence- and
        sentence-level features, yielding two tensors (one for each feature type).
    2. Concatenate the sequence- and sentence-level tensors along the sequence dimension
        by appending sentence-level features at the first available token position after
        the sequence-level (token-level) features.

    Arguments:
        attribute: Name of attribute (e.g. `text` or `label`) whose features will be
            processed.
        attribute_signature: A dictionary containing two lists of feature signatures,
            one for each feature type (`sequence` or `sentence`) of the given attribute.
        config: A model config used for correctly parameterising the layer and the
            `ConcatenateSparseDenseFeatures` layer it uses internally.

    Input shape:
        Tuple of three input tensors:
            sequence_features: List of 3-D dense or sparse tensors, each with shape
                `(batch_size, max_seq_length, input_dim)` where `input_dim` can be
                different for sparse vs dense tensors. See the input shape of
                `ConcatenateSparseDenseFeatures` for more information.
            sentence_features: List of 3-D dense or sparse tensors, each with shape
                `(batch_size, 1, input_dim)` where `input_dim` can be different for
                sparse vs dense tensors, and can differ from that in
                `sequence_features`. See the input shape of
                `ConcatenateSparseDenseFeatures` for more information.
            sequence_feature_lengths: Dense tensor of shape `(batch_size, )`.

    Output shape:
        combined_features: A 3-D tensor with shape `(batch_size, sequence_length,
            units)` where `units` is  completely  determined by the internally applied
            `ConcatenateSparseDenseFeatures` layer and `sequence_length` is the combined
            length of sequence- and sentence-level features: `max_seq_length + 1` if
            both feature types are present, `max_seq_length` if only sequence-level
            features are present, and 1 if only sentence-level features are present).
        mask_combined_sequence_sentence: A 3-D tensor with shape
            `(batch_size, sequence_length, 1)`.

    Raises:
        A `TFLayerConfigException` if no feature signatures are provided.

    Attributes:
        output_units: The last dimension size of the layer's `combined_features` output.
    """

    def __init__(
        self,
        attribute: Text,
        attribute_signature: Dict[Text, List[FeatureSignature]],
        config: Dict[Text, Any],
    ) -> None:
        """Creates a new `RasaFeatureCombiningLayer` object."""
        if not attribute_signature or not (
            attribute_signature.get(SENTENCE, [])
            or attribute_signature.get(SEQUENCE, [])
        ):
            raise TFLayerConfigException(
                "The attribute signature must contain some feature signatures."
            )

        super().__init__(name=f"rasa_feature_combining_layer_{attribute}")

        self._tf_layers: Dict[Text, tf.keras.layers.Layer] = {}

        # Prepare sparse-dense combining layers for each present feature type
        self._feature_types_present = self._get_present_feature_types(
            attribute_signature
        )
        self._prepare_sparse_dense_concat_layers(attribute, attribute_signature, config)

        # Prepare components for combining sequence- and sentence-level features
        self._prepare_sequence_sentence_concat(attribute, config)

        self.output_units = self._calculate_output_units(attribute, config)

    @staticmethod
    def _get_present_feature_types(
        attribute_signature: Dict[Text, List[FeatureSignature]],
    ) -> Dict[Text, bool]:
        """Determines feature types that are present.

        Knowing which feature types are present is important because many downstream
        operations depend on it, e.g. combining sequence- and sentence-level features
        is only done if both feature types are present.
        """
        return {
            feature_type: (
                feature_type in attribute_signature
                and len(attribute_signature[feature_type]) > 0
            )
            for feature_type in [SEQUENCE, SENTENCE]
        }

    def _prepare_sparse_dense_concat_layers(
        self,
        attribute: Text,
        attribute_signature: Dict[Text, List[FeatureSignature]],
        config: Dict[Text, Any],
    ) -> None:
        """Prepares sparse-dense combining layers for all present feature types."""
        for feature_type, present in self._feature_types_present.items():
            if not present:
                continue
            self._tf_layers[f"sparse_dense.{feature_type}"] = (
                ConcatenateSparseDenseFeatures(
                    attribute=attribute,
                    feature_type=feature_type,
                    feature_type_signature=attribute_signature[feature_type],
                    config=config,
                )
            )

    def _prepare_sequence_sentence_concat(
        self, attribute: Text, config: Dict[Text, Any]
    ) -> None:
        """Sets up combining sentence- and sequence-level features (if needed).

        This boils down to preparing for unifying the units of the sequence- and
        sentence-level features if they differ -- the same number of units is required
        for combining the features.
        """
        if (
            self._feature_types_present[SEQUENCE]
            and self._feature_types_present[SENTENCE]
        ):
            # The output units of this layer will be based on the output sizes of the
            # sparse+dense combining layers that are internally applied to all features.
            sequence_units = self._tf_layers[f"sparse_dense.{SEQUENCE}"].output_units
            sentence_units = self._tf_layers[f"sparse_dense.{SENTENCE}"].output_units

            # Last dimension needs to be unified if sequence- and sentence-level
            # features have different sizes, e.g. due to being produced by different
            # featurizers.
            if sequence_units != sentence_units:
                for feature_type in [SEQUENCE, SENTENCE]:
                    self._tf_layers[
                        f"unify_dims_before_seq_sent_concat.{feature_type}"
                    ] = layers.Ffnn(
                        layer_name_suffix=f"unify_dims.{attribute}_{feature_type}",
                        layer_sizes=[config[CONCAT_DIMENSION][attribute]],
                        dropout_rate=config[DROP_RATE],
                        reg_lambda=config[REGULARIZATION_CONSTANT],
                        density=config[CONNECTION_DENSITY],
                    )

    def _calculate_output_units(self, attribute: Text, config: Dict[Text, Any]) -> int:
        """Calculates the number of output units for this layer class.

        The number depends mainly on whether dimension unification is used or not.
        """
        # If dimension unification is used, output units are determined by the unifying
        # layers.
        if (
            f"unify_dims_before_seq_sent_concat.{SEQUENCE}" in self._tf_layers
            or f"unify_dims_before_seq_sent_concat.{SENTENCE}" in self._tf_layers
        ):
            return config[CONCAT_DIMENSION][attribute]
        # Without dimension unification, the units from the underlying sparse_dense
        # layers are carried over and should be the same for sequence-level features
        # (if present) as for sentence-level features.
        elif self._feature_types_present[SEQUENCE]:
            return self._tf_layers[f"sparse_dense.{SEQUENCE}"].output_units
        return self._tf_layers[f"sparse_dense.{SENTENCE}"].output_units

    def _concat_sequence_sentence_features(
        self,
        sequence_tensor: tf.Tensor,
        sentence_tensor: tf.Tensor,
        mask_combined_sequence_sentence: tf.Tensor,
    ) -> tf.Tensor:
        """Concatenates sequence- & sentence-level features along sequence dimension."""
        # If needed, pass both feature types through a dense layer to bring them to the
        # same shape.
        if f"unify_dims_before_seq_sent_concat.{SEQUENCE}" in self._tf_layers:
            sequence_tensor = self._tf_layers[
                f"unify_dims_before_seq_sent_concat.{SEQUENCE}"
            ](sequence_tensor)
        if f"unify_dims_before_seq_sent_concat.{SENTENCE}" in self._tf_layers:
            sentence_tensor = self._tf_layers[
                f"unify_dims_before_seq_sent_concat.{SENTENCE}"
            ](sentence_tensor)

        # mask_combined_sequence_sentence has for each input example a sequence of 1s of
        # the length seq_length+1, where seq_length is the number of real tokens. The
        # rest is 0s which form a padding up to the max. sequence length + 1 (max.
        # number of real tokens + 1). Here the mask is turned into a mask that has 0s
        # everywhere and 1 only at the immediate next position after the last real
        # token's position for a given input example. Example (batch size = 2, sequence
        # lengths = [1, 2]):
        # [[[1], [0], [0]],     ___\   [[[0], [1], [0]],
        #  [[1], [1], [0]]]        /    [[0], [0], [1]]]
        sentence_feature_positions_mask = (
            mask_combined_sequence_sentence
            * tf.math.cumprod(
                1 - mask_combined_sequence_sentence,
                axis=1,
                exclusive=True,
                reverse=True,
            )
        )

        # The new mask is used to distribute the sentence features at the sequence
        # positions marked by 1s. The sentence features' dimensionality effectively
        # changes from `(batch_size, 1, feature_dim)` to `(batch_size, max_seq_length+1,
        # feature_dim)`, but the array is sparse, with real features present only at
        # positions determined by 1s in the mask.
        sentence_tensor = sentence_feature_positions_mask * sentence_tensor

        # Padding of sequence-level features is increased by 1 in the sequence
        # dimension to match the shape of modified sentence-level features.
        sequence_tensor = tf.pad(sequence_tensor, [[0, 0], [0, 1], [0, 0]])

        # Sequence- and sentence-level features effectively get concatenated by
        # summing the two padded feature arrays like this (batch size = 1):
        # [[seq1, seq2, seq3, 0, 0]] + [[0, 0, 0, sent1, 0]] =
        # = [[seq1, seq2, seq3, sent1, 0]]
        return sequence_tensor + sentence_tensor

    def _combine_sequence_level_features(
        self,
        sequence_features: List[Union[tf.Tensor, tf.SparseTensor]],
        mask_sequence: tf.Tensor,
        training: bool,
    ) -> Optional[tf.Tensor]:
        """Processes & combines sequence-level features if any are present."""
        if self._feature_types_present[SEQUENCE]:
            sequence_features_combined = self._tf_layers[f"sparse_dense.{SEQUENCE}"](
                (sequence_features,), training=training
            )

            # Apply mask which has 1s at positions of real tokens and 0s at all padded
            # token positions. This is needed because the sparse+dense combining layer
            # might've turned some fake (padded) features (i.e. 0s) into non-zero
            # numbers and we want those to become zeros again.
            # This step isn't needed for sentence-level features because those are never
            # padded -- the effective sequence length in their case is always 1.
            return sequence_features_combined * mask_sequence

        return None

    def _combine_sentence_level_features(
        self,
        sentence_features: List[Union[tf.Tensor, tf.SparseTensor]],
        sequence_feature_lengths: tf.Tensor,
        training: bool,
    ) -> Tuple[Optional[tf.Tensor], Optional[tf.Tensor]]:
        """Processes & combines sentence-level features if any are present."""
        if self._feature_types_present[SENTENCE]:
            sentence_features_combined = self._tf_layers[f"sparse_dense.{SENTENCE}"](
                (sentence_features,), training=training
            )
            # Sentence-level features have sequence dimension of length 1, add it to
            # sequence-level feature lengths.
            combined_sequence_sentence_feature_lengths = sequence_feature_lengths + 1

        else:
            sentence_features_combined = None

            # Without sentence-level features, the feature sequence lengths are
            # completely determined by sequence-level features.
            combined_sequence_sentence_feature_lengths = sequence_feature_lengths

        return sentence_features_combined, combined_sequence_sentence_feature_lengths

    def call(
        self,
        inputs: Tuple[
            List[Union[tf.Tensor, tf.SparseTensor]],
            List[Union[tf.Tensor, tf.SparseTensor]],
            tf.Tensor,
        ],
        training: bool = False,
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """Combines multiple 3-D dense/sparse feature tensors into one.

        Arguments:
            inputs: Tuple containing:
                sequence_features: Dense or sparse tensors representing different
                    token-level features.
                sentence_features: Dense or sparse tensors representing sentence-level
                    features.
                sequence_feature_lengths: A tensor containing the real sequence length
                    (the number of real -- not padding -- tokens) for each example in
                    the batch.
            training: A flag indicating whether the layer should behave in training mode
                (applying dropout to sparse tensors if applicable) or in inference mode
                (not applying dropout).

        Returns:
            combined features: A tensor containing all the features combined.
            mask_combined_sequence_sentence: A binary mask with 1s in place of real
                features in the combined feature tensor, and 0s in padded positions with
                fake features.
        """
        sequence_features = inputs[0]
        sentence_features = inputs[1]
        sequence_feature_lengths = inputs[2]

        # This mask is specifically for sequence-level features.
        mask_sequence = compute_mask(sequence_feature_lengths)

        sequence_features_combined = self._combine_sequence_level_features(
            sequence_features, mask_sequence, training
        )

        (
            sentence_features_combined,
            combined_sequence_sentence_feature_lengths,
        ) = self._combine_sentence_level_features(
            sentence_features, sequence_feature_lengths, training
        )

        mask_combined_sequence_sentence = compute_mask(
            combined_sequence_sentence_feature_lengths
        )

        # If both feature types are present, combine them. Otherwise, just the present
        # feature type will be returned.
        if (
            sequence_features_combined is not None
            and sentence_features_combined is not None
        ):
            features_to_return = self._concat_sequence_sentence_features(
                sequence_features_combined,
                sentence_features_combined,
                mask_combined_sequence_sentence,
            )
        elif sequence_features_combined is not None:
            features_to_return = sequence_features_combined
        else:
            features_to_return = sentence_features_combined

        return features_to_return, mask_combined_sequence_sentence


class RasaSequenceLayer(RasaCustomLayer):
    """Creates an embedding from all features for a sequence attribute; facilitates MLM.

    This layer combines all features for an attribute and embeds them using a
    transformer, optionally doing masked language modeling. The layer is meant only for
    attributes with sequence-level features, such as `text`, `response` and
    `action_text`.

    Internally, this layer applies the following steps:
    1. Combine features using `RasaFeatureCombiningLayer`.
    2. Apply a dense layer(s) to the combined features.
    3. Optionally, and only during training for the `text` attribute, apply masking to
        the features and create further helper variables for masked language modeling.
    4. Embed the features using a transformer, effectively reducing variable-length
        sequences of features to fixed-size embeddings.

    Arguments:
        attribute: Name of attribute (e.g. `text` or `label`) whose features will be
            processed.
        attribute_signature: A dictionary containing two lists of feature signatures,
            one for each feature type (`sentence` or `sequence`) of the given attribute.
        config: A model config used for correctly parameterising the underlying layers.

    Input shape:
        Tuple of three input tensors:
            sequence_features: List of 3-D dense or sparse tensors, each with shape
                `(batch_size, max_seq_length, input_dim)` where `input_dim` can be
                different for sparse vs dense tensors. See the input shape of
                `ConcatenateSparseDenseFeatures` for more information.
            sentence_features: List of 3-D dense or sparse tensors, each with shape
                `(batch_size, 1, input_dim)` where `input_dim` can be different for
                sparse vs dense tensors, and can differ from that in
                `sequence_features`. See the input shape of
                `ConcatenateSparseDenseFeatures` for more information.
            sequence_feature_lengths: Dense tensor of shape `(batch_size, )`.

    Output shape:
        outputs: `(batch_size, seq_length, units)` where `units` matches the underlying
            transformer's output size (if present), otherwise it matches the output size
            of the `Ffnn` block applied to the combined features, or it's the output
            size of the underlying `RasaFeatureCombiningLayer` if the `Ffnn` block has 0
            layers. `seq_length` is the sum of the sequence dimension
            sizes of sequence- and sentence-level features (for details, see the output
            shape of `RasaFeatureCombiningLayer`). If both feature types are present,
            then `seq_length` will be 1 + the length of the longest sequence of real
            tokens across all examples in the given batch.
        seq_sent_features: `(batch_size, seq_length, hidden_dim)`, where `hidden_dim` is
            the output size of the underlying `Ffnn` block, or the output size of the
            underlying `RasaFeatureCombiningLayer` if the `Ffnn` block has 0 layers.
        mask_combined_sequence_sentence: `(batch_size, seq_length, 1)`
        token_ids: `(batch_size, seq_length, id_dim)`. `id_dim` is 2 when no dense
            sequence-level features are present. Otherwise, it's arbitrarily chosen to
            match the last dimension size of the first dense sequence-level feature in
            the input list of features.
        mlm_boolean_mask: `(batch_size, seq_length, 1)`, empty tensor if not doing MLM.
        attention_weights: `(transformer_layers, batch_size, num_transformer_heads,
            seq_length, seq_length)`, empty tensor if the transformer has 0 layers.

    Raises:
        A `TFLayerConfigException` if no feature signatures for sequence-level features
            are provided.

    Attributes:
        output_units: The last dimension size of the layer's first output (`outputs`).
    """

    FEATURE_COMBINING = "feature_combining"
    FFNN = "ffnn"
    TRANSFORMER = "transformer"
    MLM_INPUT_MASK = "mlm_input_mask"
    SPARSE_TO_DENSE_FOR_TOKEN_IDS = "sparse_to_dense_for_token_ids"

    def __init__(
        self,
        attribute: Text,
        attribute_signature: Dict[Text, List[FeatureSignature]],
        config: Dict[Text, Any],
    ) -> None:
        """Creates a new `RasaSequenceLayer` object."""
        if not attribute_signature or not attribute_signature.get(SEQUENCE, []):
            raise TFLayerConfigException(
                "The attribute signature must contain some sequence-level feature"
                "signatures but none were found."
            )

        super().__init__(name=f"rasa_sequence_layer_{attribute}")

        self._tf_layers: Dict[Text, Any] = {
            self.FEATURE_COMBINING: RasaFeatureCombiningLayer(
                attribute, attribute_signature, config
            ),
            self.FFNN: layers.Ffnn(
                config[HIDDEN_LAYERS_SIZES][attribute],
                config[DROP_RATE],
                config[REGULARIZATION_CONSTANT],
                config[CONNECTION_DENSITY],
                layer_name_suffix=attribute,
            ),
        }

        self._enables_mlm = False
        # Note: Within TED, masked language modeling becomes just input dropout,
        # since there is no loss term associated with predicting the masked tokens.
        self._prepare_masked_language_modeling(attribute, attribute_signature, config)

        transformer_layers, transformer_units = self._prepare_transformer(
            attribute, config
        )
        self._has_transformer = transformer_layers > 0

        self.output_units = self._calculate_output_units(
            attribute, transformer_layers, transformer_units, config
        )

    @staticmethod
    def _get_transformer_dimensions(
        attribute: Text, config: Dict[Text, Any]
    ) -> Tuple[int, int]:
        """Determines # of transformer layers & output size from the model config.

        The config can contain these directly (same for all attributes) or specified
        separately for each attribute.
        If a transformer is used (e.i. if `number_of_transformer_layers` is positive),
        the default `transformer_size` which is `None` breaks things. Thus,
        we need to set a reasonable default value so that the model works fine.
        """
        transformer_layers = config[NUM_TRANSFORMER_LAYERS]
        if isinstance(transformer_layers, dict):
            transformer_layers = transformer_layers[attribute]
        transformer_units = config[TRANSFORMER_SIZE]
        if isinstance(transformer_units, dict):
            transformer_units = transformer_units[attribute]
        if transformer_layers > 0 and (not transformer_units or transformer_units < 1):
            transformer_units = DEFAULT_TRANSFORMER_SIZE

        return transformer_layers, transformer_units

    def _prepare_transformer(
        self, attribute: Text, config: Dict[Text, Any]
    ) -> Tuple[int, int]:
        """Creates a transformer & returns its number of layers and output units."""
        transformer_layers, transformer_units = self._get_transformer_dimensions(
            attribute, config
        )
        self._tf_layers[self.TRANSFORMER] = prepare_transformer_layer(
            attribute_name=attribute,
            config=config,
            num_layers=transformer_layers,
            units=transformer_units,
            drop_rate=config[DROP_RATE],
            unidirectional=config[UNIDIRECTIONAL_ENCODER],
        )
        return transformer_layers, transformer_units

    def _prepare_masked_language_modeling(
        self,
        attribute: Text,
        attribute_signature: Dict[Text, List[FeatureSignature]],
        config: Dict[Text, Any],
    ) -> None:
        """Prepares masking and computing helper variables for masked language modeling.

        Only done for the text attribute and only if sequence-level (token-level)
        features are present (MLM requires token-level information).
        """
        if attribute == TEXT and SEQUENCE in attribute_signature and config[MASKED_LM]:
            self._enables_mlm = True
            self._tf_layers[self.MLM_INPUT_MASK] = layers.InputMask()

            # Unique IDs of different token types are needed to construct the possible
            # label space for MLM. If dense features are present, they're used as such
            # IDs, othwerise sparse features are embedded by a non-trainable
            # DenseForSparse layer to create small embeddings that serve as IDs.
            expect_dense_seq_features = any(
                [not signature.is_sparse for signature in attribute_signature[SEQUENCE]]
            )
            if not expect_dense_seq_features:
                self._tf_layers[self.SPARSE_TO_DENSE_FOR_TOKEN_IDS] = (
                    layers.DenseForSparse(
                        units=2,
                        use_bias=False,
                        trainable=False,
                        name=f"{self.SPARSE_TO_DENSE_FOR_TOKEN_IDS}.{attribute}",
                    )
                )

    def _calculate_output_units(
        self,
        attribute: Text,
        transformer_layers: int,
        transformer_units: int,
        config: Dict[Text, Any],
    ) -> int:
        """Determines the output units based on what layer components are present.

        The size depends on which component is the last created one in the internal
        pipeline that is `RasaFeatureCombiningLayer` -> `Ffnn` -> `Transformer`, since
        not all the components are always created.
        """
        # transformer is the last component
        if transformer_layers > 0:
            return transformer_units

        # the Ffnn block is the last component
        if len(config[HIDDEN_LAYERS_SIZES][attribute]) > 0:
            # this is the output size of the last layer of the Ffnn block
            return config[HIDDEN_LAYERS_SIZES][attribute][-1]

        # only the RasaFeatureCombiningLayer is present
        return self._tf_layers[self.FEATURE_COMBINING].output_units

    def _features_as_token_ids(
        self, features: List[Union[tf.Tensor, tf.SparseTensor]]
    ) -> Optional[tf.Tensor]:
        """Creates dense labels (token IDs) used for negative sampling in MLM."""
        # If there are dense features, we use them as labels - taking the first dense
        # feature in the list, but any other dense feature would do the job.
        for f in features:
            if not isinstance(f, tf.SparseTensor):
                return tf.stop_gradient(f)

        # If no dense features are found, use a sparse feature but convert it into
        # a dense one first.
        for f in features:
            if isinstance(f, tf.SparseTensor):
                return tf.stop_gradient(
                    self._tf_layers[self.SPARSE_TO_DENSE_FOR_TOKEN_IDS](f)
                )

        return None

    def _create_mlm_tensors(
        self,
        sequence_features: List[Union[tf.Tensor, tf.SparseTensor]],
        seq_sent_features: tf.Tensor,
        mask_sequence: tf.Tensor,
        sentence_features_present: bool,
        training: bool,
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        """Produces helper variables for masked language modelling (only in training).

        The `token_ids` embeddings can be viewed as token-level labels/unique IDs of all
        input tokens (to be used later in the MLM loss) because these embeddings aren't
        affected by dropout or masking and are effectively always unique for different
        input tokens (and same for the same tokens).
        `token_ids` share the batch and sequence dimension with the combined sequence-
        and sentence-level features, the last dimension is unimportant and mimics the
        first dense sequence-level feature in the list of features, or alternatively the
        last dimension will have size 2 if there are only sparse sequence features
        present.
        """
        token_ids = self._features_as_token_ids(sequence_features)

        # Pad in the sequence dimension to match the shape of combined sequence- and
        # sentence-level features. This means padding by 1 if sentence-level features
        # are present (those effectively have sequence length of 1) and not padding
        # otherwise.
        if sentence_features_present:
            token_ids = tf.pad(token_ids, [[0, 0], [0, 1], [0, 0]])
            mask_sequence = tf.pad(mask_sequence, [[0, 0], [0, 1], [0, 0]])

        # mlm_boolean_mask has the same shape as the tensor with all combined features
        # (except the last dimension), with True meaning tokens that are masked and
        # False meaning tokens that aren't masked or that are fake (padded) tokens.
        # Note that only sequence-level features are masked, nothing happens to the
        # sentence-level features in the combined features tensor.
        seq_sent_features, mlm_boolean_mask = self._tf_layers[self.MLM_INPUT_MASK](
            seq_sent_features, mask_sequence, training=training
        )

        return seq_sent_features, token_ids, mlm_boolean_mask

    def call(
        self,
        inputs: Tuple[
            List[Union[tf.Tensor, tf.SparseTensor]],
            List[Union[tf.Tensor, tf.SparseTensor]],
            tf.Tensor,
        ],
        training: bool = False,
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        """Combines all of an attribute's features and embeds using a transformer.

        Arguments:
            inputs: Tuple containing:
                sequence_features: Dense or sparse tensors representing different
                    token-level features.
                sentence_features: Dense or sparse tensors representing different
                    sentence-level features.
                sequence_feature_lengths: A tensor containing the real sequence length
                    (the number of real -- not padding -- tokens) for each example in
                    the batch.
            training: A flag indicating whether the layer should behave in training mode
                (applying dropout to sparse tensors if applicable) or in inference mode
                (not applying dropout).

        Returns:
            outputs: Tensor with all features combined, masked (if doing MLM) and
                embedded with a transformer.
            seq_sent_features: Tensor with all features combined from just before the
                masking and transformer is applied
            mask_combined_sequence_sentence: A binary mask with 1s in place of real
                features in the combined feature tensor, and 0s in padded positions with
                fake features.
            token_ids: Tensor with dense token-level features which can serve as
                IDs (unique embeddings) of all the different tokens found in the batch.
                Empty tensor if not doing MLM.
            mlm_boolean_mask: A boolean mask with `True` where real tokens in `outputs`
                were masked and `False` elsewhere. Empty tensor if not doing MLM.
            attention_weights: Tensor containing self-attention weights received
                from the underlying transformer. Empty tensor if the transformer has 0
                layers.
        """
        sequence_features = inputs[0]
        sentence_features = inputs[1]
        sequence_feature_lengths = inputs[2]

        # Combine all features (sparse/dense, sequence-/sentence-level) into one tensor,
        # also get a binary mask that has 1s at positions with real features and 0s at
        # padded positions.
        seq_sent_features, mask_combined_sequence_sentence = self._tf_layers[
            self.FEATURE_COMBINING
        ]((sequence_features, sentence_features, sequence_feature_lengths))

        # Apply one or more dense layers.
        seq_sent_features = self._tf_layers[self.FFNN](
            seq_sent_features, training=training
        )

        # If using masked language modeling, mask the transformer inputs and get labels
        # for the masked tokens and a boolean mask. Note that TED does not use MLM loss,
        # hence using masked language modeling (if enabled) becomes just input dropout.
        if self._enables_mlm and training:
            mask_sequence = compute_mask(sequence_feature_lengths)
            (
                seq_sent_features_masked,
                token_ids,
                mlm_boolean_mask,
            ) = self._create_mlm_tensors(
                sequence_features,
                seq_sent_features,
                mask_sequence,
                sentence_features_present=len(sentence_features) > 0,
                training=training,
            )
        else:
            # tf.zeros((0,)) is an alternative to None
            token_ids = tf.zeros((0,))
            mlm_boolean_mask = tf.zeros((0,))
            seq_sent_features_masked = seq_sent_features

        # Apply the transformer (if present), hence reducing a sequences of features per
        # input example into a simple fixed-size embedding.
        if self._has_transformer:
            mask_padding = 1 - mask_combined_sequence_sentence
            outputs, attention_weights = self._tf_layers[self.TRANSFORMER](
                seq_sent_features_masked, mask_padding, training=training
            )
            outputs = tf.nn.gelu(outputs)
        else:
            # tf.zeros((0,)) is an alternative to None
            outputs, attention_weights = seq_sent_features_masked, tf.zeros((0,))

        return (
            outputs,
            seq_sent_features,
            mask_combined_sequence_sentence,
            token_ids,
            mlm_boolean_mask,
            attention_weights,
        )


def compute_mask(sequence_lengths: tf.Tensor) -> tf.Tensor:
    """Computes binary mask given real sequence lengths.

    Takes a 1-D tensor of shape `(batch_size,)` containing the lengths of sequences
    (in terms of number of tokens) in the batch. Creates a binary mask of shape
    `(batch_size, max_seq_length, 1)` with 1s at positions with real tokens and 0s
    elsewhere.
    """
    mask = tf.sequence_mask(sequence_lengths, dtype=tf.float32)
    return tf.expand_dims(mask, -1)


def prepare_transformer_layer(
    attribute_name: Text,
    config: Dict[Text, Any],
    num_layers: int,
    units: int,
    drop_rate: float,
    unidirectional: bool,
) -> Union[
    TransformerEncoder,
    Callable[
        [tf.Tensor, Optional[tf.Tensor], Optional[Union[tf.Tensor, bool]]],
        Tuple[tf.Tensor, Optional[tf.Tensor]],
    ],
]:
    """Creates & returns a transformer encoder, potentially with 0 layers."""
    if num_layers > 0:
        return TransformerEncoder(
            num_layers,
            units,
            config[NUM_HEADS],
            units * 4,
            config[REGULARIZATION_CONSTANT],
            dropout_rate=drop_rate,
            attention_dropout_rate=config[DROP_RATE_ATTENTION],
            density=config[CONNECTION_DENSITY],
            unidirectional=unidirectional,
            use_key_relative_position=config[KEY_RELATIVE_ATTENTION],
            use_value_relative_position=config[VALUE_RELATIVE_ATTENTION],
            max_relative_position=config[MAX_RELATIVE_POSITION],
            name=f"{attribute_name}_encoder",
        )
    # create lambda so that it can be used later without the check
    return lambda x, mask, training: (x, None)



from __future__ import annotations

import logging
from abc import abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Text,
    Tuple,
    Type,
    Union,
    cast,
)

import numpy as np
from tqdm import tqdm

import rasa.shared.core.trackers
import rasa.shared.utils.io
from rasa.core.exceptions import InvalidTrackerFeaturizerUsageError
from rasa.core.featurizers.precomputation import MessageContainerForCoreFeaturization
from rasa.core.featurizers.single_state_featurizer import SingleStateFeaturizer
from rasa.shared.core.constants import (
    ACTION_UNLIKELY_INTENT_NAME,
    PREVIOUS_ACTION,
    USER,
)
from rasa.shared.core.domain import Domain, State
from rasa.shared.core.events import ActionExecuted, Event, UserUttered
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import ACTION_NAME, ENTITIES, INTENT, TEXT
from rasa.shared.nlu.training_data.features import Features
from rasa.utils.tensorflow.constants import LABEL_PAD_ID
from rasa.utils.tensorflow.model_data import ragged_array_to_ndarray

FEATURIZER_FILE = "featurizer.json"

logger = logging.getLogger(__name__)


class InvalidStory(RasaException):
    """Exception that can be raised if story cannot be featurized."""

    def __init__(self, message: Text) -> None:
        """Creates an InvalidStory exception.

        Args:
          message: a custom exception message.
        """
        self.message = message
        super(InvalidStory, self).__init__(message)

    def __str__(self) -> Text:
        return self.message


class TrackerFeaturizer:
    """Base class for actual tracker featurizers."""

    # Class registry to store all subclasses
    _registry: ClassVar[Dict[str, Type["TrackerFeaturizer"]]] = {}
    _featurizer_type: str = "TrackerFeaturizer"

    def __init__(
        self, state_featurizer: Optional[SingleStateFeaturizer] = None
    ) -> None:
        """Initializes the tracker featurizer.

        Args:
            state_featurizer: The state featurizer used to encode tracker states.
        """
        self.state_featurizer = state_featurizer

    @classmethod
    def register(cls, featurizer_type: str) -> Callable:
        """Decorator to register featurizer subclasses."""

        def wrapper(subclass: Type["TrackerFeaturizer"]) -> Type["TrackerFeaturizer"]:
            cls._registry[featurizer_type] = subclass
            # Store the type identifier in the class for serialization
            subclass._featurizer_type = featurizer_type
            return subclass

        return wrapper

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TrackerFeaturizer":
        """Create featurizer instance from dictionary."""
        featurizer_type = data.pop("type")

        if featurizer_type not in cls._registry:
            raise ValueError(f"Unknown featurizer type: {featurizer_type}")

        # Get the correct subclass and instantiate it
        subclass = cls._registry[featurizer_type]
        return subclass.create_from_dict(data)

    @classmethod
    @abstractmethod
    def create_from_dict(cls, data: Dict[str, Any]) -> "TrackerFeaturizer":
        """Each subclass must implement its own creation from dict method."""
        pass

    @staticmethod
    def _create_states(
        tracker: DialogueStateTracker,
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
    ) -> List[State]:
        """Creates states for the given tracker.

        Args:
            tracker: The tracker to transform to states.
            domain: The domain of the tracker.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_rule_only_turns: If `True` ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.

        Returns:
            Trackers as states.
        """
        return tracker.past_states(
            domain,
            omit_unset_slots=omit_unset_slots,
            ignore_rule_only_turns=ignore_rule_only_turns,
            rule_only_data=rule_only_data,
        )

    def _featurize_states(
        self,
        trackers_as_states: List[List[State]],
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> List[List[Dict[Text, List[Features]]]]:
        """Featurizes state histories with `state_featurizer`.

        Args:
            trackers_as_states: Lists of states produced by a `DialogueStateTracker`
                instance.
            precomputations: Contains precomputed features and attributes.

        Returns:
            Featurized tracker states.
        """
        if self.state_featurizer is None:
            return [[{}]]
        else:
            return [
                [
                    self.state_featurizer.encode_state(state, precomputations)
                    for state in tracker_states
                ]
                for tracker_states in trackers_as_states
            ]

    @staticmethod
    def _convert_labels_to_ids(
        trackers_as_actions: List[List[Text]], domain: Domain
    ) -> np.ndarray:
        """Converts actions to label ids for each tracker.

        Args:
            trackers_as_actions: A list of tracker labels.
            domain: The domain containing action names.

        Returns:
            Label IDs for each tracker
        """
        # store labels in numpy arrays so that it corresponds to np arrays of input
        # features
        return ragged_array_to_ndarray(
            [
                np.array(
                    [domain.index_for_action(action) for action in tracker_actions]
                )
                for tracker_actions in trackers_as_actions
            ]
        )

    def _create_entity_tags(
        self,
        trackers_as_entities: List[List[Dict[Text, Any]]],
        precomputations: Optional[MessageContainerForCoreFeaturization],
        bilou_tagging: bool = False,
    ) -> List[List[Dict[Text, List[Features]]]]:
        """Featurizes extracted entities with `state_featurizer`.

        Args:
            trackers_as_entities: Extracted entities from trackers.
            precomputations: Contains precomputed features and attributes.
            bilou_tagging: When `True` use the BILOU tagging scheme.

        Returns:
            Trackers as entity features.
        """
        if self.state_featurizer is None:
            return [[{}]]
        else:
            return [
                [
                    self.state_featurizer.encode_entities(
                        entity_data, precomputations, bilou_tagging
                    )
                    for entity_data in trackers_entities
                ]
                for trackers_entities in trackers_as_entities
            ]

    @staticmethod
    def _entity_data(event: UserUttered) -> Dict[Text, Any]:
        """Extracts entities from event if not using intents.

        Args:
            event: The event from which to extract entities.

        Returns:
            Event text and entities if no intent is present.
        """
        # train stories support both text and intent,
        # but if intent is present, the text is ignored
        if event.text and not event.intent_name:
            return {TEXT: event.text, ENTITIES: event.entities}

        # input is not textual, so add empty dict
        return {}

    @staticmethod
    def _remove_user_text_if_intent(trackers_as_states: List[List[State]]) -> None:
        """Deletes user text from state dictionaries if intent is present.

        Only featurizing either the intent or user text is currently supported. When
        both are present in a state, the user text is removed so that only the intent
        is featurized.

        `trackers_as_states` is modified in place.

        Args:
            trackers_as_states: States produced by a `DialogueStateTracker` instance.
        """
        for states in trackers_as_states:
            for state in states:
                # remove text features to only use intent
                if state.get(USER, {}).get(INTENT) and state.get(USER, {}).get(TEXT):
                    del state[USER][TEXT]

    def training_states_and_labels(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[List[List[State]], List[List[Text]]]:
        """Transforms trackers to states and labels.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            Trackers as states and labels.
        """
        (
            trackers_as_states,
            trackers_as_labels,
            _,
        ) = self.training_states_labels_and_entities(
            trackers,
            domain,
            omit_unset_slots=omit_unset_slots,
            ignore_action_unlikely_intent=ignore_action_unlikely_intent,
        )
        return trackers_as_states, trackers_as_labels

    @abstractmethod
    def training_states_labels_and_entities(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[List[List[State]], List[List[Text]], List[List[Dict[Text, Any]]]]:
        """Transforms trackers to states, labels, and entity data.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            Trackers as states, labels, and entity data.
        """
        raise NotImplementedError(
            f"`{self.__class__.__name__}` should implement how to "
            f"encode trackers as feature vectors"
        )

    def prepare_for_featurization(
        self, domain: Domain, bilou_tagging: bool = False
    ) -> None:
        """Ensures that the featurizer is ready to be called during training.

        State featurizer needs to build its vocabulary from the domain
        for it to be ready to be used during training.

        Args:
            domain: Domain of the assistant.
            bilou_tagging: Whether to consider bilou tagging.
        """
        if self.state_featurizer is None:
            raise InvalidTrackerFeaturizerUsageError(
                f"Instance variable 'state_featurizer' is not set. "
                f"During initialization set 'state_featurizer' to an instance of "
                f"'{SingleStateFeaturizer.__class__.__name__}' class "
                f"to get numerical features for trackers."
            )
        self.state_featurizer.prepare_for_training(domain, bilou_tagging)

    def featurize_trackers(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        precomputations: Optional[MessageContainerForCoreFeaturization],
        bilou_tagging: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[
        List[List[Dict[Text, List[Features]]]],
        np.ndarray,
        List[List[Dict[Text, List[Features]]]],
    ]:
        """Featurizes the training trackers.

        Args:
            trackers: list of training trackers
            domain: the domain
            precomputations: Contains precomputed features and attributes.
            bilou_tagging: indicates whether BILOU tagging should be used or not
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training state features.

        Returns:
            - a dictionary of state types (INTENT, TEXT, ACTION_NAME, ACTION_TEXT,
              ENTITIES, SLOTS, ACTIVE_LOOP) to a list of features for all dialogue
              turns in all training trackers
            - the label ids (e.g. action ids) for every dialogue turn in all training
              trackers
            - A dictionary of entity type (ENTITY_TAGS) to a list of features
              containing entity tag ids for text user inputs otherwise empty dict
              for all dialogue turns in all training trackers
        """
        self.prepare_for_featurization(domain, bilou_tagging)
        (
            trackers_as_states,
            trackers_as_labels,
            trackers_as_entities,
        ) = self.training_states_labels_and_entities(
            trackers,
            domain,
            ignore_action_unlikely_intent=ignore_action_unlikely_intent,
        )

        tracker_state_features = self._featurize_states(
            trackers_as_states, precomputations
        )

        if not tracker_state_features and not trackers_as_labels:
            # If input and output were empty, it means there is
            # no data on which the policy can be trained
            # hence return them as it is. They'll be handled
            # appropriately inside the policy.
            return tracker_state_features, np.ndarray(trackers_as_labels), []

        label_ids = self._convert_labels_to_ids(trackers_as_labels, domain)

        entity_tags = self._create_entity_tags(
            trackers_as_entities, precomputations, bilou_tagging
        )

        return tracker_state_features, label_ids, entity_tags

    def _choose_last_user_input(
        self, trackers_as_states: List[List[State]], use_text_for_last_user_input: bool
    ) -> None:
        for states in trackers_as_states:
            last_state = states[-1]
            # only update the state of the real user utterance
            if not rasa.shared.core.trackers.is_prev_action_listen_in_state(last_state):
                continue

            if use_text_for_last_user_input:
                # remove intent features to only use text
                if last_state.get(USER, {}).get(INTENT):
                    del last_state[USER][INTENT]
                # don't add entities if text is used for featurization
                if last_state.get(USER, {}).get(ENTITIES):
                    del last_state[USER][ENTITIES]
            else:
                # remove text features to only use intent
                if last_state.get(USER, {}).get(TEXT):
                    del last_state[USER][TEXT]

        # make sure that all dialogue steps are either intent or text based
        self._remove_user_text_if_intent(trackers_as_states)

    def prediction_states(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        use_text_for_last_user_input: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        ignore_action_unlikely_intent: bool = False,
    ) -> List[List[State]]:
        """Transforms trackers to states for prediction.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            use_text_for_last_user_input: Indicates whether to use text or intent label
                for featurizing last user input.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.
            ignore_action_unlikely_intent: Whether to remove states containing
                `action_unlikely_intent` from prediction states.

        Returns:
            Trackers as states for prediction.
        """
        raise NotImplementedError(
            "Featurizer must have the capacity to create feature vector"
        )

    def create_state_features(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        precomputations: Optional[MessageContainerForCoreFeaturization],
        use_text_for_last_user_input: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        ignore_action_unlikely_intent: bool = False,
    ) -> List[List[Dict[Text, List[Features]]]]:
        """Creates state features for prediction.

        Args:
            trackers: A list of state trackers
            domain: The domain
            precomputations: Contains precomputed features and attributes.
            use_text_for_last_user_input: Indicates whether to use text or intent label
                for featurizing last user input.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.
            ignore_action_unlikely_intent: Whether to remove any states containing
                `action_unlikely_intent` from state features.

        Returns:
            Dictionaries of state type (INTENT, TEXT, ACTION_NAME, ACTION_TEXT,
            ENTITIES, SLOTS, ACTIVE_LOOP) to a list of features for all dialogue
            turns in all trackers.
        """
        trackers_as_states = self.prediction_states(
            trackers,
            domain,
            use_text_for_last_user_input,
            ignore_rule_only_turns,
            rule_only_data,
            ignore_action_unlikely_intent=ignore_action_unlikely_intent,
        )
        return self._featurize_states(trackers_as_states, precomputations)

    def persist(self, path: Union[Text, Path]) -> None:
        """Persists the tracker featurizer to the given path.

        Args:
            path: The path to persist the tracker featurizer to.
        """
        featurizer_file = Path(path) / FEATURIZER_FILE
        rasa.shared.utils.io.create_directory_for_file(featurizer_file)

        # entity tags are persisted in TED policy, they are not needed for prediction
        if self.state_featurizer is not None:
            self.state_featurizer.entity_tag_specs = []

        # noinspection PyTypeChecker
        rasa.shared.utils.io.dump_obj_as_json_to_file(featurizer_file, self.to_dict())

    @staticmethod
    def load(path: Union[Text, Path]) -> Optional[TrackerFeaturizer]:
        """Loads the featurizer from file.

        Args:
            path: The path to load the tracker featurizer from.

        Returns:
            The loaded tracker featurizer.
        """
        featurizer_file = Path(path) / FEATURIZER_FILE
        if featurizer_file.is_file():
            data = rasa.shared.utils.io.read_json_file(featurizer_file)

            if "type" not in data:
                logger.error(
                    f"Couldn't load featurizer for policy. "
                    f"File '{featurizer_file}' does not contain all "
                    f"necessary information. 'type' is missing."
                )
                return None

            return TrackerFeaturizer.from_dict(data)

        logger.error(
            f"Couldn't load featurizer for policy. "
            f"File '{featurizer_file}' doesn't exist."
        )
        return None

    @staticmethod
    def _remove_action_unlikely_intent_from_states(states: List[State]) -> List[State]:
        return [
            state
            for state in states
            if not _is_prev_action_unlikely_intent_in_state(state)
        ]

    @staticmethod
    def _remove_action_unlikely_intent_from_events(events: List[Event]) -> List[Event]:
        return [
            event
            for event in events
            if (
                not isinstance(event, ActionExecuted)
                or event.action_name != ACTION_UNLIKELY_INTENT_NAME
            )
        ]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.__class__._featurizer_type,
            "state_featurizer": (
                self.state_featurizer.to_dict() if self.state_featurizer else None
            ),
        }


@TrackerFeaturizer.register("FullDialogueTrackerFeaturizer")
class FullDialogueTrackerFeaturizer(TrackerFeaturizer):
    """Creates full dialogue training data for time distributed architectures.

    Creates training data that uses each time output for prediction.
    """

    def training_states_labels_and_entities(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[List[List[State]], List[List[Text]], List[List[Dict[Text, Any]]]]:
        """Transforms trackers to states, action labels, and entity data.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            Trackers as states, action labels, and entity data.
        """
        trackers_as_states = []
        trackers_as_actions = []
        trackers_as_entities = []

        logger.debug(
            "Creating states and action examples from "
            "collected trackers (by {}({}))..."
            "".format(type(self).__name__, type(self.state_featurizer).__name__)
        )
        pbar = tqdm(
            trackers,
            desc="Processed trackers",
            disable=rasa.shared.utils.io.is_logging_disabled(),
        )
        for tracker in pbar:
            states = self._create_states(
                tracker, domain, omit_unset_slots=omit_unset_slots
            )
            events = tracker.applied_events(True)

            if ignore_action_unlikely_intent:
                states = self._remove_action_unlikely_intent_from_states(states)
                events = self._remove_action_unlikely_intent_from_events(events)

            delete_first_state = False
            actions = []
            entities = []
            entity_data = {}
            for event in events:
                if isinstance(event, UserUttered):
                    entity_data = self._entity_data(event)

                if not isinstance(event, ActionExecuted):
                    continue

                if not event.unpredictable:
                    # only actions which can be
                    # predicted at a stories start
                    action = event.action_name or event.action_text
                    if action is not None:
                        actions.append(action)
                    entities.append(entity_data)
                else:
                    # unpredictable actions can be
                    # only the first in the story
                    if delete_first_state:
                        raise InvalidStory(
                            f"Found two unpredictable actions in one story "
                            f"'{tracker.sender_id}'. Check your story files."
                        )
                    delete_first_state = True

                # reset entity_data for the the next turn
                entity_data = {}

            if delete_first_state:
                states = states[1:]

            trackers_as_states.append(states[:-1])
            trackers_as_actions.append(actions)
            trackers_as_entities.append(entities)

        self._remove_user_text_if_intent(trackers_as_states)

        return trackers_as_states, trackers_as_actions, trackers_as_entities

    def prediction_states(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        use_text_for_last_user_input: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        ignore_action_unlikely_intent: bool = False,
    ) -> List[List[State]]:
        """Transforms trackers to states for prediction.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            use_text_for_last_user_input: Indicates whether to use text or intent label
                for featurizing last user input.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.
            ignore_action_unlikely_intent: Whether to remove any states containing
                `action_unlikely_intent` from prediction states.

        Returns:
            Trackers as states for prediction.
        """
        trackers_as_states = [
            self._create_states(
                tracker,
                domain,
                ignore_rule_only_turns=ignore_rule_only_turns,
                rule_only_data=rule_only_data,
            )
            for tracker in trackers
        ]

        if ignore_action_unlikely_intent:
            trackers_as_states = [
                self._remove_action_unlikely_intent_from_states(states)
                for states in trackers_as_states
            ]

        self._choose_last_user_input(trackers_as_states, use_text_for_last_user_input)

        return trackers_as_states

    def to_dict(self) -> Dict[str, Any]:
        return super().to_dict()

    @classmethod
    def create_from_dict(cls, data: Dict[str, Any]) -> "FullDialogueTrackerFeaturizer":
        state_featurizer = SingleStateFeaturizer.create_from_dict(
            data["state_featurizer"]
        )
        return cls(
            state_featurizer,
        )


@TrackerFeaturizer.register("MaxHistoryTrackerFeaturizer")
class MaxHistoryTrackerFeaturizer(TrackerFeaturizer):
    """Truncates the tracker history into `max_history` long sequences.

    Creates training data from trackers where actions are the output prediction
    labels. Tracker state sequences which represent policy input are truncated
    to not excede `max_history` states.
    """

    LABEL_NAME = "action"

    def __init__(
        self,
        state_featurizer: Optional[SingleStateFeaturizer] = None,
        max_history: Optional[int] = None,
        remove_duplicates: bool = True,
    ) -> None:
        """Initializes the tracker featurizer.

        Args:
            state_featurizer: The state featurizer used to encode the states.
            max_history: The maximum length of an extracted state sequence.
            remove_duplicates: Keep only unique training state sequence/label pairs.
        """
        super().__init__(state_featurizer)
        self.max_history = max_history
        self.remove_duplicates = remove_duplicates

    @staticmethod
    def slice_state_history(
        states: List[State], slice_length: Optional[int]
    ) -> List[State]:
        """Slices states from the trackers history.

        Args:
            states: The states
            slice_length: The slice length

        Returns:
            The sliced states.
        """
        if not slice_length:
            return states

        return states[-slice_length:]

    @staticmethod
    def _hash_example(states: List[State], labels: Optional[List[Text]] = None) -> int:
        """Hashes states (and optionally label).

        Produces a hash of the tracker state sequence (and optionally the labels).
        If `labels` is `None`, labels don't get hashed.

        Args:
            states: The tracker state sequence to hash.
            labels: Label strings associated with this state sequence.

        Returns:
            The hash of the states and (optionally) the label.
        """
        frozen_states = tuple(
            s if s is None else DialogueStateTracker.freeze_current_state(s)
            for s in states
        )
        if labels is not None:
            frozen_labels = tuple(labels)
            return hash((frozen_states, frozen_labels))
        else:
            return hash(frozen_states)

    def training_states_labels_and_entities(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[List[List[State]], List[List[Text]], List[List[Dict[Text, Any]]]]:
        """Transforms trackers to states, action labels, and entity data.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            Trackers as states, labels, and entity data.
        """
        example_states = []
        example_labels = []
        example_entities = []

        # Store of example hashes for removing duplicate training examples.
        hashed_examples = set()

        logger.debug(
            f"Creating states and {self.LABEL_NAME} label examples from "
            f"collected trackers "
            f"(by {type(self).__name__}({type(self.state_featurizer).__name__}))..."
        )
        pbar = tqdm(
            trackers,
            desc="Processed trackers",
            disable=rasa.shared.utils.io.is_logging_disabled(),
        )
        for tracker in pbar:
            for states, label, entities in self._extract_examples(
                tracker,
                domain,
                omit_unset_slots=omit_unset_slots,
                ignore_action_unlikely_intent=ignore_action_unlikely_intent,
            ):
                if self.remove_duplicates:
                    hashed = self._hash_example(states, label)
                    if hashed in hashed_examples:
                        continue
                    hashed_examples.add(hashed)

                example_states.append(states)
                example_labels.append(label)
                example_entities.append(entities)

                pbar.set_postfix({f"# {self.LABEL_NAME}": f"{len(example_labels):d}"})

        self._remove_user_text_if_intent(example_states)

        logger.debug(f"Created {len(example_states)} {self.LABEL_NAME} examples.")

        return example_states, example_labels, example_entities

    def _extract_examples(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Iterator[Tuple[List[State], List[Text], List[Dict[Text, Any]]]]:
        """Creates an iterator over training examples from a tracker.

        Args:
            tracker: The tracker from which to extract training examples.
            domain: The domain of the training data.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            An iterator over example states, labels, and entity data.
        """
        tracker_states = self._create_states(
            tracker, domain, omit_unset_slots=omit_unset_slots
        )
        events = tracker.applied_events(True)

        if ignore_action_unlikely_intent:
            tracker_states = self._remove_action_unlikely_intent_from_states(
                tracker_states
            )
            events = self._remove_action_unlikely_intent_from_events(events)

        label_index = 0
        entity_data = {}
        for event in events:
            if isinstance(event, UserUttered):
                entity_data = self._entity_data(event)

            elif isinstance(event, ActionExecuted):
                label_index += 1

                # use only actions which can be predicted at a stories start
                if event.unpredictable:
                    continue

                sliced_states = self.slice_state_history(
                    tracker_states[:label_index], self.max_history
                )
                label = cast(List[Text], [event.action_name or event.action_text])
                entities = [entity_data]

                yield sliced_states, label, entities

                # reset entity_data for the the next turn
                entity_data = {}

    def prediction_states(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        use_text_for_last_user_input: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        ignore_action_unlikely_intent: bool = False,
    ) -> List[List[State]]:
        """Transforms trackers to states for prediction.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            use_text_for_last_user_input: Indicates whether to use text or intent label
                for featurizing last user input.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.
            ignore_action_unlikely_intent: Whether to remove any states containing
                `action_unlikely_intent` from prediction states.

        Returns:
            Trackers as states for prediction.
        """
        trackers_as_states = [
            self._create_states(
                tracker,
                domain,
                ignore_rule_only_turns=ignore_rule_only_turns,
                rule_only_data=rule_only_data,
            )
            for tracker in trackers
        ]

        # Remove `action_unlikely_intent` from `trackers_as_states`.
        # This must be done before state history slicing to ensure the
        # max history of the sliced states matches training time.
        if ignore_action_unlikely_intent:
            trackers_as_states = [
                self._remove_action_unlikely_intent_from_states(states)
                for states in trackers_as_states
            ]

        trackers_as_states = [
            self.slice_state_history(states, self.max_history)
            for states in trackers_as_states
        ]
        self._choose_last_user_input(trackers_as_states, use_text_for_last_user_input)

        return trackers_as_states

    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()
        data.update(
            {
                "remove_duplicates": self.remove_duplicates,
                "max_history": self.max_history,
            }
        )
        return data

    @classmethod
    def create_from_dict(cls, data: Dict[str, Any]) -> "MaxHistoryTrackerFeaturizer":
        state_featurizer = SingleStateFeaturizer.create_from_dict(
            data["state_featurizer"]
        )
        return cls(state_featurizer, data["max_history"], data["remove_duplicates"])


@TrackerFeaturizer.register("IntentMaxHistoryTrackerFeaturizer")
class IntentMaxHistoryTrackerFeaturizer(MaxHistoryTrackerFeaturizer):
    """Truncates the tracker history into `max_history` long sequences.

    Creates training data from trackers where intents are the output prediction
    labels. Tracker state sequences which represent policy input are truncated
    to not excede `max_history` states.
    """

    LABEL_NAME = "intent"

    @classmethod
    def _convert_labels_to_ids(
        cls, trackers_as_intents: List[List[Text]], domain: Domain
    ) -> np.ndarray:
        """Converts a list of labels to a matrix of label ids.

        The number of rows is equal to `len(trackers_as_intents)`. The number of
        columns is equal to the maximum number of positive labels that any training
        example is associated with. Rows are padded with `LABEL_PAD_ID` if not all rows
        have the same number of labels.

        Args:
            trackers_as_intents: Positive example label ids
                associated with each training example.
            domain: The domain of the training data.

        Returns:
           A matrix of label ids.
        """
        # store labels in numpy arrays so that it corresponds to np arrays
        # of input features
        label_ids = [
            [domain.intents.index(intent) for intent in tracker_intents]
            for tracker_intents in trackers_as_intents
        ]

        return np.array(cls._pad_label_ids(label_ids))

    @staticmethod
    def _pad_label_ids(label_ids: List[List[int]]) -> List[List[int]]:
        """Pads label ids so that all are of the same length.

        Args:
            label_ids: Label ids of varying lengths

        Returns:
            Label ids padded to be of uniform length.
        """
        # If `label_ids` is an empty list, no padding needs to be added.
        if not label_ids:
            return label_ids

        # Add `LABEL_PAD_ID` padding to labels array so that
        # each example has equal number of labels
        multiple_labels_count = [len(a) for a in label_ids]
        max_labels_count = max(multiple_labels_count)
        num_padding_needed = [max_labels_count - len(a) for a in label_ids]

        padded_label_ids = []
        for ids, num_pads in zip(label_ids, num_padding_needed):
            padded_row = list(ids) + [LABEL_PAD_ID] * num_pads
            padded_label_ids.append(padded_row)
        return padded_label_ids

    def training_states_labels_and_entities(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Tuple[List[List[State]], List[List[Text]], List[List[Dict[Text, Any]]]]:
        """Transforms trackers to states, intent labels, and entity data.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            Trackers as states, labels, and entity data.
        """
        example_states = []
        example_entities = []

        # Store of example hashes (of both states and labels) for removing
        # duplicate training examples.
        hashed_examples = set()
        # Mapping of example state hash to set of
        # positive labels associated with the state.
        state_hash_to_label_set: DefaultDict[int, Set[Text]] = defaultdict(set)

        logger.debug(
            f"Creating states and {self.LABEL_NAME} label examples from "
            f"collected trackers "
            f"(by {type(self).__name__}({type(self.state_featurizer).__name__}))..."
        )
        pbar = tqdm(
            trackers,
            desc="Processed trackers",
            disable=rasa.shared.utils.io.is_logging_disabled(),
        )
        for tracker in pbar:
            for states, label, entities in self._extract_examples(
                tracker,
                domain,
                omit_unset_slots=omit_unset_slots,
                ignore_action_unlikely_intent=ignore_action_unlikely_intent,
            ):
                if self.remove_duplicates:
                    hashed = self._hash_example(states, label)
                    if hashed in hashed_examples:
                        continue
                    hashed_examples.add(hashed)

                # Store all positive labels associated with a training state.
                state_hash = self._hash_example(states)

                # Only add unique example states unless `remove_duplicates` is `False`.
                if (
                    not self.remove_duplicates
                    or state_hash not in state_hash_to_label_set
                ):
                    example_states.append(states)
                    example_entities.append(entities)

                state_hash_to_label_set[state_hash].add(label[0])

                pbar.set_postfix({f"# {self.LABEL_NAME}": f"{len(example_states):d}"})

        # Collect positive labels for each state example.
        example_labels = [
            list(state_hash_to_label_set[self._hash_example(state)])
            for state in example_states
        ]

        self._remove_user_text_if_intent(example_states)

        logger.debug(f"Created {len(example_states)} {self.LABEL_NAME} examples.")

        return example_states, example_labels, example_entities

    def _extract_examples(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        omit_unset_slots: bool = False,
        ignore_action_unlikely_intent: bool = False,
    ) -> Iterator[Tuple[List[State], List[Text], List[Dict[Text, Any]]]]:
        """Creates an iterator over training examples from a tracker.

        Args:
            tracker: The tracker from which to extract training examples.
            domain: The domain of the training data.
            omit_unset_slots: If `True` do not include the initial values of slots.
            ignore_action_unlikely_intent: Whether to remove `action_unlikely_intent`
                from training states.

        Returns:
            An iterator over example states, labels, and entity data.
        """
        tracker_states = self._create_states(
            tracker, domain, omit_unset_slots=omit_unset_slots
        )
        events = tracker.applied_events(True)

        if ignore_action_unlikely_intent:
            tracker_states = self._remove_action_unlikely_intent_from_states(
                tracker_states
            )
            events = self._remove_action_unlikely_intent_from_events(events)

        label_index = 0
        for event in events:
            if isinstance(event, ActionExecuted):
                label_index += 1

            elif isinstance(event, UserUttered):
                sliced_states = self.slice_state_history(
                    tracker_states[:label_index], self.max_history
                )
                label = cast(List[Text], [event.intent_name or event.text])
                entities: List[Dict[Text, Any]] = [{}]

                yield sliced_states, label, entities

    @staticmethod
    def _cleanup_last_user_state_with_action_listen(
        trackers_as_states: List[List[State]],
    ) -> List[List[State]]:
        """Removes the last tracker state if the previous action is `action_listen`.

        States with the previous action equal to `action_listen` correspond to states
        with a new user intent. This information is what `UnexpecTEDIntentPolicy` is
        trying to predict so it needs to be removed before obtaining a prediction.

        Args:
            trackers_as_states: Trackers converted to states

        Returns:
            Filtered states with last `action_listen` removed.
        """
        for states in trackers_as_states:
            if not states:
                continue
            last_state = states[-1]
            if rasa.shared.core.trackers.is_prev_action_listen_in_state(last_state):
                del states[-1]

        return trackers_as_states

    def prediction_states(
        self,
        trackers: List[DialogueStateTracker],
        domain: Domain,
        use_text_for_last_user_input: bool = False,
        ignore_rule_only_turns: bool = False,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        ignore_action_unlikely_intent: bool = False,
    ) -> List[List[State]]:
        """Transforms trackers to states for prediction.

        Args:
            trackers: The trackers to transform.
            domain: The domain.
            use_text_for_last_user_input: Indicates whether to use text or intent label
                for featurizing last user input.
            ignore_rule_only_turns: If True ignore dialogue turns that are present
                only in rules.
            rule_only_data: Slots and loops,
                which only occur in rules but not in stories.
            ignore_action_unlikely_intent: Whether to remove any states containing
                `action_unlikely_intent` from prediction states.

        Returns:
            Trackers as states for prediction.
        """
        trackers_as_states = [
            self._create_states(
                tracker,
                domain,
                ignore_rule_only_turns=ignore_rule_only_turns,
                rule_only_data=rule_only_data,
            )
            for tracker in trackers
        ]

        # Remove `action_unlikely_intent` from `trackers_as_states`.
        # This must be done before state history slicing to ensure the
        # max history of the sliced states matches training time.
        if ignore_action_unlikely_intent:
            trackers_as_states = [
                self._remove_action_unlikely_intent_from_states(states)
                for states in trackers_as_states
            ]

        self._choose_last_user_input(trackers_as_states, use_text_for_last_user_input)

        # `tracker_as_states` contain a state with intent = last intent
        # and previous action = action_listen. This state needs to be
        # removed as it was not present during training as well because
        # predicting the last intent is what the policies using this
        # featurizer do. This is specifically done before state history
        # is sliced so that the number of past states is same as `max_history`.
        self._cleanup_last_user_state_with_action_listen(trackers_as_states)

        trackers_as_states = [
            self.slice_state_history(states, self.max_history)
            for states in trackers_as_states
        ]

        return trackers_as_states

    def to_dict(self) -> Dict[str, Any]:
        return super().to_dict()

    @classmethod
    def create_from_dict(
        cls, data: Dict[str, Any]
    ) -> "IntentMaxHistoryTrackerFeaturizer":
        state_featurizer = SingleStateFeaturizer.create_from_dict(
            data["state_featurizer"]
        )
        return cls(state_featurizer, data["max_history"], data["remove_duplicates"])


def _is_prev_action_unlikely_intent_in_state(state: State) -> bool:
    prev_action_name = state.get(PREVIOUS_ACTION, {}).get(ACTION_NAME)
    return prev_action_name == ACTION_UNLIKELY_INTENT_NAME



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text

import structlog
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders.directory import DirectoryLoader
from langchain_community.document_loaders.text import TextLoader
from langchain_community.vectorstores.faiss import FAISS

from rasa.core.information_retrieval import (
    InformationRetrieval,
    InformationRetrievalException,
    SearchResultList,
)
from rasa.core.information_retrieval.ingestion.faq_parser import _format_faq_documents
from rasa.utils.endpoints import EndpointConfig
from rasa.utils.ml_utils import persist_faiss_vector_store

if TYPE_CHECKING:
    from langchain.schema import Document
    from langchain.schema.embeddings import Embeddings

logger = structlog.get_logger()


class FAISS_Store(InformationRetrieval):
    """FAISS Store implementation."""

    def __init__(
        self,
        embeddings: "Embeddings",
        index_path: str,
        docs_folder: Optional[str],
        create_index: Optional[bool] = False,
        parse_as_faq_pairs: Optional[bool] = False,
    ):
        """Initializes the FAISS Store."""
        self.chunk_size = 1000
        self.chunk_overlap = 20
        self.parse_as_faq_pairs = parse_as_faq_pairs

        path = Path(index_path) / "documents_faiss"
        if create_index:
            logger.info(
                "information_retrieval.faiss_store.create_index", path=path.absolute()
            )
            self.index = self._create_document_index(docs_folder, embeddings)
            self._persist(path)
        else:
            logger.info(
                "information_retrieval.faiss_store.load_index", path=path.absolute()
            )
            self.index = FAISS.load_local(
                str(path), embeddings, allow_dangerous_deserialization=True
            )

    @staticmethod
    def load_documents(docs_folder: str) -> List["Document"]:
        """Loads documents from a given folder.

        Args:
            docs_folder: The folder containing the documents.

        Returns:
            the list of documents
        """
        logger.info(
            "information_retrieval.faiss_store.load_documents",
            docs_folder=Path(docs_folder).absolute(),
        )
        loader = DirectoryLoader(
            docs_folder, glob="**/*.txt", loader_cls=TextLoader, show_progress=True
        )

        return loader.load()

    def _create_document_index(
        self, docs_folder: Optional[str], embedding: "Embeddings"
    ) -> FAISS:
        """Creates a document index from the documents in the given folder.
        Args:
            docs_folder: The folder containing the documents.
            embedding: The embedding to use.
        Returns:
            The document index.
        """
        if not docs_folder:
            raise ValueError("parameter `docs_folder` needs to be specified")

        documents = self.load_documents(docs_folder)

        if not self.parse_as_faq_pairs:
            splitter = RecursiveCharacterTextSplitter(
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
                length_function=len,
            )
            parsed_documents = splitter.split_documents(documents)
        else:
            parsed_documents = _format_faq_documents(documents)

        logger.info(
            "information_retrieval.faiss_store._create_document_index",
            len_chunks=len(parsed_documents),
        )

        if not parsed_documents:
            raise ValueError(f"No documents found at '{docs_folder}'.")

        texts = [document.page_content for document in parsed_documents]
        metadatas = [document.metadata for document in parsed_documents]

        # Warn if document size exceeds 2GB - FAISS in-memory performance degrades
        ONE_GB = 1024 * 1024 * 1024
        total_size_bytes = sum(len(text.encode("utf-8")) for text in texts)
        size_limit_bytes = 2 * ONE_GB  # 2GB
        if total_size_bytes > size_limit_bytes:
            logger.warning(
                "information_retrieval.faiss_store.large_document_size",
                message="Document size exceeds 2GB. FAISS in-memory vector store "
                "performance may degrade significantly with large datasets. "
                "Consider using a dedicated vector database for better performance.",
                total_size_gb=round(total_size_bytes / ONE_GB, 2),
            )

        # Batch size to avoid hitting embedding API token limits
        # Use a conservative batch size for large documents
        batch_size = 50
        total_batches = (len(texts) + batch_size - 1) // batch_size

        # Create index with first batch
        logger.info(
            "information_retrieval.faiss_store._create_document_index.embedding_batch",
            batch=1,
            total_batches=total_batches,
        )
        index = FAISS.from_texts(
            texts[:batch_size], embedding, metadatas=metadatas[:batch_size]
        )

        # Add remaining documents in batches
        for i in range(batch_size, len(texts), batch_size):
            batch_num = (i // batch_size) + 1
            logger.info(
                "information_retrieval.faiss_store._create_document_index.embedding_batch",
                batch=batch_num,
                total_batches=total_batches,
            )
            batch_texts = texts[i : i + batch_size]
            batch_metadatas = metadatas[i : i + batch_size]
            index.add_texts(batch_texts, metadatas=batch_metadatas)

        return index

    def _persist(self, path: Path) -> None:
        persist_faiss_vector_store(path, self.index)

    def connect(self, config: EndpointConfig) -> None:
        """Faiss does not need to connect to a server."""
        pass

    async def search(
        self, query: Text, tracker_state: Dict[str, Any], threshold: float = 0.0
    ) -> SearchResultList:
        logger.debug("information_retrieval.faiss_store.search", query=query)
        try:
            documents = await self.index.as_retriever().ainvoke(query)
        except Exception as exc:
            raise InformationRetrievalException from exc

        return SearchResultList.from_document_list(documents)



from typing import Any, Dict, List, Optional, Union

import structlog
from pydantic import BaseModel, Field, field_validator

from rasa.builder.copilot.models import (
    ChatMessage,
    CopilotContext,
    EventContent,
    ReferenceEntry,
    ResponseCategory,
    create_chat_message_from_dict,
)
from rasa.builder.document_retrieval.models import Document
from rasa.builder.shared.tracker_context import TrackerContext

structlogger = structlog.get_logger()


class DatasetInput(BaseModel):
    """Model for the input field of a dataset entry."""

    message: Optional[str] = None
    tracker_event_attachments: List[EventContent] = Field(default_factory=list)


class DatasetExpectedOutput(BaseModel):
    """Model for the expected_output field of a dataset entry."""

    answer: str
    response_category: ResponseCategory
    references: list[ReferenceEntry]


class DatasetMetadataCopilotAdditionalContext(BaseModel):
    """Model for the copilot_additional_context in metadata."""

    relevant_documents: List[Document] = Field(default_factory=list)
    relevant_assistant_files: Dict[str, str] = Field(default_factory=dict)
    assistant_tracker_context: Optional[Dict[str, Any]] = None
    assistant_logs: str = Field(default="")
    copilot_chat_history: List[ChatMessage] = Field(default_factory=list)

    @field_validator("copilot_chat_history", mode="before")
    @classmethod
    def parse_chat_history(
        cls, v: Union[List[Dict[str, Any]], List[ChatMessage]]
    ) -> List[ChatMessage]:
        """Manually parse chat history messages based on role field."""
        # If already parsed ChatMessage objects, return them as-is
        if (
            v
            and isinstance(v, list)
            and all(isinstance(item, ChatMessage) for item in v)
        ):
            return v  # type: ignore[return-value]

        # Check for mixed types (some ChatMessage, some not)
        if (
            v
            and isinstance(v, list)
            and any(isinstance(item, ChatMessage) for item in v)
        ):
            message = (
                "Mixed types in copilot_chat_history: cannot mix ChatMessage objects "
                "with other types."
            )
            structlogger.error(
                "dataset_entry.parse_chat_history.mixed_types",
                event_info=message,
                chat_history_types=[type(item) for item in v],
            )
            raise ValueError(message)

        # Otherwise, parse from dictionaries
        parsed_messages: List[ChatMessage] = []
        for message_data in v:
            chat_message = create_chat_message_from_dict(message_data)
            parsed_messages.append(chat_message)
        return parsed_messages


class DatasetMetadata(BaseModel):
    """Model for the metadata field of a dataset entry."""

    ids: Dict[str, str] = Field(default_factory=dict)
    copilot_additional_context: DatasetMetadataCopilotAdditionalContext = Field(
        default_factory=DatasetMetadataCopilotAdditionalContext
    )


class DatasetEntry(BaseModel):
    """Pydantic model for dataset entries from Langfuse ExperimentItem."""

    # Basic fields from ExperimentItem
    id: str
    input: DatasetInput
    expected_output: DatasetExpectedOutput
    metadata: DatasetMetadata

    def to_copilot_context(self) -> CopilotContext:
        """Create a CopilotContext from the dataset entry.

        Raises:
            ValueError: If the metadata is None, as it's required for creating a valid
                CopilotContext.

        Returns:
            CopilotContext with all the context information.
        """
        if self.metadata is None:
            message = (
                f"Cannot create CopilotContext from dataset item with id: {self.id}. "
                f"Metadata is required but was None."
            )
            structlogger.error(
                "dataset_entry.to_copilot_context.metadata_is_none",
                event_info=message,
                item_id=self.id,
                item_metadata=self.metadata,
            )
            raise ValueError(message)

        # Parse tracker context if available
        tracker_context = None
        if (
            self.metadata.copilot_additional_context.assistant_tracker_context
            is not None
        ):
            tracker_context = TrackerContext(
                **self.metadata.copilot_additional_context.assistant_tracker_context
            )

        return CopilotContext(
            tracker_context=tracker_context,
            assistant_logs=self.metadata.copilot_additional_context.assistant_logs,
            assistant_files=self.metadata.copilot_additional_context.relevant_assistant_files,
            copilot_chat_history=self.metadata.copilot_additional_context.copilot_chat_history,
        )

    @classmethod
    def from_raw_data(
        cls,
        id: str,
        input_data: Dict[str, Any],
        expected_output_data: Dict[str, Any],
        metadata_data: Dict[str, Any],
    ) -> "DatasetEntry":
        """Create a DatasetEntry from raw dictionary data.

        Args:
            id: The dataset entry ID.
            input_data: Raw input dictionary.
            expected_output_data: Raw expected output dictionary.
            metadata_data: Raw metadata dictionary with all the additional context
                used to generate the Copilot response.

        Returns:
            DatasetEntry with parsed data.
        """
        # Use Pydantic's model_validate to parse nested structures
        dataset_input = DatasetInput.model_validate(input_data)
        dataset_expected_output = DatasetExpectedOutput.model_validate(
            expected_output_data
        )
        dataset_metadata = DatasetMetadata.model_validate(metadata_data)

        return cls(
            id=id,
            input=dataset_input,
            expected_output=dataset_expected_output,
            metadata=dataset_metadata,
        )



from rasa.agents.protocol.a2a.a2a_agent import A2AAgent
from rasa.agents.protocol.mcp.mcp_open_agent import MCPOpenAgent
from rasa.agents.protocol.mcp.mcp_task_agent import MCPTaskAgent

__all__ = ["A2AAgent", "MCPOpenAgent", "MCPTaskAgent"]



DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256

LOG_COMPONENT_SOURCE_METHOD_INIT = "init"
LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON = "fingerprint_addon"


LANGFUSE_METADATA_SESSION_ID = "session_id"
LANGFUSE_METADATA_AGENT_ID = "agent_id"
LANGFUSE_METADATA_MODEL_ID = "model_id"
LANGFUSE_METADATA_COMPONENT_NAME = "component_name"
LANGFUSE_METADATA_TAGS = "tags"
LANGFUSE_METADATA_CUSTOM_METADATA = "trace_metadata"
LANGFUSE_METADATA_REACT_SUB_AGENT_NAME = "react_sub_agent_name"



class EvaluationError(Exception):
    """Base exception for evaluation-related errors."""

    pass


class ClaimExtractionError(Exception):
    """Raised when a claim extraction error occurs."""

    pass



import os
import threading
import time
from typing import Dict, Optional, Tuple
from urllib.parse import ParseResult, urlencode, urlunparse

import boto3
import redis
import structlog
from aws_msk_iam_sasl_signer import MSKAuthTokenProvider
from botocore.exceptions import BotoCoreError
from botocore.model import ServiceId
from botocore.session import get_session
from botocore.signers import RequestSigner
from cachetools import TTLCache, cached

from rasa.core.constants import (
    ELASTICACHE_REDIS_AWS_IAM_ENABLED_ENV_VAR_NAME,
    KAFKA_MSK_AWS_IAM_ENABLED_ENV_VAR_NAME,
    KAFKA_SERVICE_NAME,
    RDS_SQL_DB_AWS_IAM_ENABLED_ENV_VAR_NAME,
    REDIS_SERVICE_NAME,
    SQL_SERVICE_NAME,
)
from rasa.core.iam_credentials_providers.credentials_provider_protocol import (
    IAMCredentialsProvider,
    IAMCredentialsProviderInput,
    SupportedServiceType,
    TemporaryCredentials,
)
from rasa.shared.exceptions import ConnectionException

structlogger = structlog.get_logger(__name__)

SERVICE_CONFIG: Dict[Tuple[SupportedServiceType, str], str] = {
    (
        SupportedServiceType.TRACKER_STORE,
        SQL_SERVICE_NAME,
    ): RDS_SQL_DB_AWS_IAM_ENABLED_ENV_VAR_NAME,
    (
        SupportedServiceType.TRACKER_STORE,
        REDIS_SERVICE_NAME,
    ): ELASTICACHE_REDIS_AWS_IAM_ENABLED_ENV_VAR_NAME,
    (
        SupportedServiceType.EVENT_BROKER,
        KAFKA_SERVICE_NAME,
    ): KAFKA_MSK_AWS_IAM_ENABLED_ENV_VAR_NAME,
    (
        SupportedServiceType.LOCK_STORE,
        REDIS_SERVICE_NAME,
    ): ELASTICACHE_REDIS_AWS_IAM_ENABLED_ENV_VAR_NAME,
}


class AWSRDSIAMCredentialsProvider(IAMCredentialsProvider):
    """Generates temporary credentials for AWS RDS using IAM roles."""

    def __init__(self, username: str, host: str, port: int) -> None:
        """Initializes the provider."""
        self.username = username
        self.host = host
        self.port = port

    def get_temporary_credentials(self) -> TemporaryCredentials:
        """Generates temporary credentials for AWS RDS."""
        structlogger.debug(
            "rasa.core.aws_rds_iam_credentials_provider.get_credentials",
            event_info="IAM authentication for AWS RDS enabled. "
            "Generating temporary auth token...",
        )

        try:
            client = boto3.client("rds")
            auth_token = client.generate_db_auth_token(
                DBHostname=self.host,
                Port=self.port,
                DBUsername=self.username,
            )
            structlogger.info(
                "rasa.core.aws_rds_iam_credentials_provider.generated_credentials",
                event_info="Successfully generated temporary auth token for AWS RDS.",
            )
            return TemporaryCredentials(auth_token=auth_token)
        except (BotoCoreError, ValueError) as exc:
            structlogger.error(
                "rasa.core.aws_rds_iam_credentials_provider.error_generating_credentials",
                event_info="Failed to generate temporary auth token for AWS RDS.",
                error=str(exc),
            )
            return TemporaryCredentials(auth_token=None)


class AWSMSKafkaIAMCredentialsProvider(IAMCredentialsProvider):
    """Generates temporary credentials for AWS MSK using IAM roles."""

    def __init__(self) -> None:
        self.region = os.getenv("AWS_DEFAULT_REGION", os.getenv("AWS_REGION"))
        self._token: Optional[str] = None
        self._expires_at: float = 0
        self.refresh_margin_seconds = 60  # Refresh 60 seconds before expiry
        # ensure thread safety when refreshing token because the
        # kafka client library we use (confluent-kafka) is multithreaded
        self.lock = threading.Lock()

    @property
    def token(self) -> Optional[str]:
        return self._token

    @token.setter
    def token(self, value: str) -> None:
        self._token = value

    @property
    def expires_at(self) -> float:
        return self._expires_at

    @expires_at.setter
    def expires_at(self, value: float) -> None:
        self._expires_at = value

    def get_temporary_credentials(self) -> TemporaryCredentials:
        """Generates temporary credentials for AWS MSK."""
        with self.lock:
            current_time = time.time()  # Current time in seconds
            if (
                not self.token
                or current_time >= self.expires_at - self.refresh_margin_seconds
            ):
                try:
                    auth_token, expiry_ms = MSKAuthTokenProvider.generate_auth_token(
                        self.region
                    )
                    structlogger.debug(
                        "rasa.core.aws_msk_iam_credentials_provider.get_credentials",
                        event_info="Successfully generated AWS IAM token for "
                        "Kafka authentication.",
                    )
                    self.token = auth_token
                    self.expires_at = int(expiry_ms) / 1000  # Convert ms to seconds
                    return TemporaryCredentials(
                        auth_token=auth_token,
                        expiration=self.expires_at,
                    )
                except Exception as exc:
                    raise ConnectionException(
                        f"Failed to generate AWS IAM token "
                        f"for MSK authentication. Original exception: {exc}"
                    ) from exc
            else:
                structlogger.debug(
                    "rasa.core.aws_msk_iam_credentials_provider.get_credentials",
                    event_info="Using cached AWS IAM token for Kafka authentication.",
                )
                return TemporaryCredentials(
                    auth_token=self.token,
                    expiration=self.expires_at,
                )


class AWSElasticacheRedisIAMCredentialsProvider(redis.CredentialProvider):
    """Generates temporary credentials for AWS ElastiCache Redis using IAM roles."""

    def __init__(self, username: str, cluster_name: Optional[str] = None) -> None:
        """Initializes the provider."""
        self.username = username
        self.cluster_name = cluster_name
        self.region = os.getenv("AWS_DEFAULT_REGION", os.getenv("AWS_REGION"))
        self.session = get_session()
        self.request_signer = RequestSigner(
            ServiceId("elasticache"),
            self.region,
            "elasticache",
            "v4",
            self.session.get_credentials(),
            self.session.get_component("event_emitter"),
        )

    # Generated IAM tokens are valid for 15 minutes
    @cached(cache=TTLCache(maxsize=128, ttl=900))
    def get_credentials(self) -> Tuple[str, str]:
        """Generates temporary credentials for AWS ElastiCache Redis.

        Required method implementation by redis-py CredentialProvider parent class.
        Used internally by redis-py when connecting to Redis.
        """
        query_params = {"Action": "connect", "User": self.username}
        url = urlunparse(
            ParseResult(
                scheme="https",
                netloc=self.cluster_name,
                path="/",
                query=urlencode(query_params),
                params="",
                fragment="",
            )
        )
        signed_url = self.request_signer.generate_presigned_url(
            {"method": "GET", "url": url, "body": {}, "headers": {}, "context": {}},
            operation_name="connect",
            expires_in=900,
            region_name=self.region,
        )

        # RequestSigner only seems to work if the URL has a protocol, but
        # Elasticache only accepts the URL without a protocol
        # So strip it off the signed URL before returning
        return self.username, signed_url.removeprefix("https://")

    def get_temporary_credentials(self) -> TemporaryCredentials:
        """Generates temporary credentials for AWS ElastiCache Redis.

        Implemented to comply with the IAMCredentialsProvider rasa-pro interface.
        Calls the get_credentials method which is used internally by redis-py.
        """
        try:
            username, signed_url = self.get_credentials()
            structlogger.info(
                "rasa.core.aws_elasticache_redis_iam_credentials_provider.generated_credentials",
                event_info="Successfully generated temporary credentials for "
                "AWS ElastiCache Redis.",
            )
            return TemporaryCredentials(username=username, presigned_url=signed_url)
        except Exception as exc:
            structlogger.error(
                "rasa.core.aws_elasticache_redis_iam_credentials_provider.error_generating_credentials",
                event_info="Failed to generate temporary credentials for "
                "AWS ElastiCache Redis.",
                error=str(exc),
            )
            return TemporaryCredentials()


def is_iam_enabled(provider_input: "IAMCredentialsProviderInput") -> bool:
    """Checks if IAM authentication is enabled for the given service."""
    service_type = provider_input.service_type
    service_name = provider_input.service_name
    iam_enabled_env_var_name = SERVICE_CONFIG.get((service_type, service_name))

    if not iam_enabled_env_var_name:
        structlogger.warning(
            "rasa.core.aws_iam_credentials_providers.is_iam_enabled.unsupported_service",
            event_info=f"IAM authentication check requested for unsupported service: "
            f"{service_name}",
        )
        return False

    return os.getenv(iam_enabled_env_var_name, "false").lower() == "true"


def create_aws_iam_credentials_provider(
    provider_input: "IAMCredentialsProviderInput",
) -> Optional["IAMCredentialsProvider"]:
    """Factory function to create an AWS IAM credentials provider."""
    iam_enabled = is_iam_enabled(provider_input)
    if not iam_enabled:
        structlogger.debug(
            "rasa.core.aws_iam_credentials_providers.create_provider.iam_not_enabled",
            event_info=f"IAM authentication not enabled for service: "
            f"{provider_input.service_type}",
        )
        return None

    if (
        provider_input.service_type == SupportedServiceType.TRACKER_STORE
        and provider_input.service_name == SQL_SERVICE_NAME
    ):
        return AWSRDSIAMCredentialsProvider(
            username=provider_input.username,
            host=provider_input.host,
            port=provider_input.port,
        )

    if (
        provider_input.service_type == SupportedServiceType.TRACKER_STORE
        and provider_input.service_name == REDIS_SERVICE_NAME
    ):
        return AWSElasticacheRedisIAMCredentialsProvider(
            username=provider_input.username,
            cluster_name=provider_input.cluster_name,
        )

    if provider_input.service_type == SupportedServiceType.EVENT_BROKER:
        return AWSMSKafkaIAMCredentialsProvider()

    if provider_input.service_type == SupportedServiceType.LOCK_STORE:
        return AWSElasticacheRedisIAMCredentialsProvider(
            username=provider_input.username,
            cluster_name=provider_input.cluster_name,
        )

    return None



import logging
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional, Text

from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage

if TYPE_CHECKING:
    from twilio.rest.api.v2010.account.message import MessageInstance

logger = logging.getLogger(__name__)


class TwilioOutput(Client, OutputChannel):
    """Output channel for Twilio."""

    @classmethod
    def name(cls) -> Text:
        return "twilio"

    def __init__(
        self,
        account_sid: Optional[Text],
        auth_token: Optional[Text],
        twilio_number: Optional[Text],
    ) -> None:
        Client.__init__(self, account_sid, auth_token)
        self.twilio_number = twilio_number
        self.send_retry = 0
        self.max_retry = 5

    async def _send_message(self, message_data: Dict[Text, Any]) -> "MessageInstance":
        message = None
        try:
            while not message and self.send_retry < self.max_retry:
                message = self.messages.create(**message_data)
                self.send_retry += 1
        except TwilioRestException as e:
            logger.error("Something went wrong " + repr(e.msg))
        finally:
            self.send_retry = 0

        if not message and self.send_retry == self.max_retry:
            logger.error("Failed to send message. Max number of retires exceeded.")

        return message

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Sends text message."""
        message_data = {"to": recipient_id, "from_": self.twilio_number}
        for message_part in text.strip().split("\n\n"):
            message_data.update({"body": message_part})
            await self._send_message(message_data)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image."""
        message_data = {
            "to": recipient_id,
            "from_": self.twilio_number,
            "media_url": [image],
        }
        await self._send_message(message_data)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Send custom json dict."""
        json_message.setdefault("to", recipient_id)
        if not json_message.get("media_url"):
            json_message.setdefault("body", "")
        if not json_message.get("messaging_service_sid"):
            json_message.setdefault("from_", self.twilio_number)

        await self._send_message(json_message)


class TwilioInput(InputChannel):
    """Twilio input channel."""

    @classmethod
    def name(cls) -> Text:
        return "twilio"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()

        return cls(
            credentials.get("account_sid"),
            credentials.get("auth_token"),
            credentials.get("twilio_number"),
        )

    @classmethod
    def _is_location_message(cls, request: Request) -> bool:
        """Check if the users message is a location."""
        return request.form.get("Latitude") and request.form.get("Longitude")

    def __init__(
        self,
        account_sid: Optional[Text],
        auth_token: Optional[Text],
        twilio_number: Optional[Text],
        debug_mode: bool = True,
    ) -> None:
        self.account_sid = account_sid
        self.auth_token = auth_token
        self.twilio_number = twilio_number
        self.debug_mode = debug_mode

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        twilio_webhook = Blueprint("twilio_webhook", __name__)

        @twilio_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @twilio_webhook.route("/webhook", methods=["POST"])
        async def message(request: Request) -> HTTPResponse:
            sender = request.form.get("From", None)
            text = request.form.get("Body", None)
            out_channel = self.get_output_channel()

            if self._is_location_message(request):
                # Text is always None with a location message/media from Twilio whatsapp
                text = "/locationData{{'Latitude': {lat},'Longitude': {long}}}".format(
                    lat=request.form.get("Latitude"), long=request.form.get("Longitude")
                )

            if sender is not None and message is not None:
                metadata = self.get_metadata(request)
                try:
                    # @ signs get corrupted in SMSes by some carriers
                    text = text.replace("¡", "@")
                    await on_new_message(
                        UserMessage(
                            text,
                            out_channel,
                            sender,
                            input_channel=self.name(),
                            metadata=metadata,
                        )
                    )
                except Exception as e:
                    logger.error(f"Exception when trying to handle message.{e}")
                    logger.debug(e, exc_info=True)
                    if self.debug_mode:
                        raise
                    pass
            else:
                logger.debug("Invalid message")

            return response.text("", status=204)

        return twilio_webhook

    def get_output_channel(self) -> OutputChannel:
        return TwilioOutput(self.account_sid, self.auth_token, self.twilio_number)



import os
from dataclasses import dataclass
from typing import AsyncIterator, Dict, Optional
from urllib.parse import urlencode

import aiohttp
import orjson
import structlog
from aiohttp import ClientTimeout, WSMsgType

from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.tts.tts_engine import (
    TTSEngine,
    TTSEngineConfig,
    TTSError,
)
from rasa.shared.constants import DEEPGRAM_API_KEY_ENV_VAR
from rasa.shared.exceptions import ConnectionException

structlogger = structlog.get_logger()


@dataclass
class DeepgramTTSConfig(TTSEngineConfig):
    model_id: Optional[str] = None
    endpoint: Optional[str] = None


class DeepgramTTS(TTSEngine[DeepgramTTSConfig]):
    session: Optional[aiohttp.ClientSession] = None
    required_env_vars = (DEEPGRAM_API_KEY_ENV_VAR,)
    ws: Optional[aiohttp.ClientWebSocketResponse] = None
    streaming_input: bool = True

    def __init__(self, config: Optional[DeepgramTTSConfig] = None):
        super().__init__(config)
        timeout = ClientTimeout(total=self.config.timeout)
        # Have to create this class-shared session lazily at run time otherwise
        # the async event loop doesn't work
        if self.__class__.session is None or self.__class__.session.closed:
            self.__class__.session = aiohttp.ClientSession(timeout=timeout)

    @staticmethod
    def get_request_headers(config: DeepgramTTSConfig) -> dict[str, str]:
        deepgram_api_key = os.environ[DEEPGRAM_API_KEY_ENV_VAR]
        return {
            "Authorization": f"Token {deepgram_api_key!s}",
        }

    async def connect(self, config: Optional[DeepgramTTSConfig] = None) -> None:
        headers = self.get_request_headers(self.config)
        ws_url = self.get_websocket_url(self.config)

        if self.session is None:
            raise ConnectionException("Client session is not initialized")

        self.ws = await self.session.ws_connect(
            ws_url,
            headers=headers,
            timeout=float(self.config.timeout),
        )

    async def close_connection(self) -> None:
        """Close WebSocket connection if it exists."""
        if self.ws and not self.ws.closed:
            await self.ws.close()
            self.ws = None

    def get_websocket_url(self, config: DeepgramTTSConfig) -> str:
        """Build WebSocket URL with query parameters."""
        base_url = config.endpoint
        query_params = {
            "model": config.model_id,
            "encoding": "mulaw",
            "sample_rate": "8000",
        }
        return f"{base_url}?{urlencode(query_params)}"

    async def send_text_chunk(self, text: str) -> None:
        """Send text to TTS engine for continuous streaming.

        This sends text to Deepgram but doesn't return anything.
        Audio will be available via stream_audio().
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.ws.send_json(
            {
                "type": "Speak",
                "text": text,
            }
        )

    async def signal_text_done(self) -> None:
        """Signal TTS engine to flush any buffered text.

        This tells Deepgram to process any remaining text and finish.
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.ws.send_json({"type": "Flush"})

    async def stream_audio(self) -> AsyncIterator[RasaAudioBytes]:
        """Stream audio output from the TTS engine.

        This continuously yields audio chunks as they arrive from Deepgram.
        Stops when it receives a Flushed or Close message.
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        try:
            async for msg in self.ws:
                if msg.type == WSMsgType.BINARY:
                    # Binary data is the raw audio - yield it
                    yield self.engine_bytes_to_rasa_audio_bytes(msg.data)

                elif msg.type == WSMsgType.TEXT:
                    # Handle control messages
                    data = orjson.loads(msg.data)
                    if data.get("type") == "Flushed":
                        # All audio has been sent, stop streaming
                        structlogger.debug("deepgram.stream_audio.flushed")
                        break
                    elif data.get("type") == "Close":
                        # Connection closing
                        structlogger.debug("deepgram.stream_audio.close")
                        break

                elif msg.type == WSMsgType.CLOSED:
                    structlogger.debug("deepgram.stream_audio.ws_closed")
                    break

                elif msg.type == WSMsgType.ERROR:
                    structlogger.error(
                        "deepgram.stream_audio.ws_error", error=str(msg.data)
                    )
                    raise TTSError(f"WebSocket error: {msg.data}")

        except Exception as e:
            structlogger.error("deepgram.stream_audio.error", error=str(e))
            raise TTSError(f"Error during audio streaming: {e}")

    async def synthesize(
        self, text: str, config: Optional[DeepgramTTSConfig] = None
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate speech from text using a remote TTS system."""
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.send_text_chunk(text)
        await self.signal_text_done()

        async for audio_chunk in self.stream_audio():
            yield audio_chunk

    def engine_bytes_to_rasa_audio_bytes(self, chunk: bytes) -> RasaAudioBytes:
        """Convert the generated tts audio bytes into rasa audio bytes."""
        # WebSocket returns raw audio bytes directly
        return RasaAudioBytes(chunk)

    @staticmethod
    def get_default_config() -> DeepgramTTSConfig:
        return DeepgramTTSConfig(
            model_id="aura-2-andromeda-en",
            endpoint="wss://api.deepgram.com/v1/speak",
            timeout=30,
        )

    @classmethod
    def from_config_dict(cls, config: Dict) -> "DeepgramTTS":
        return cls(DeepgramTTSConfig.from_dict(config))



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, cast

import structlog
from tqdm import tqdm

from rasa.core.agent import Agent
from rasa.core.channels import UserMessage
from rasa.dialogue_understanding.commands.prompt_command import PromptCommand
from rasa.dialogue_understanding.utils import set_record_commands_and_prompts
from rasa.llm_fine_tuning.conversations import Conversation, ConversationStep
from rasa.llm_fine_tuning.storage import StorageContext
from rasa.llm_fine_tuning.utils import (
    commands_as_string,
    make_mock_invoke_llm,
    patch_invoke_llm_in_generators,
)
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import KEY_USER_PROMPT, PROMPTS
from rasa.shared.utils.llm import generate_sender_id

LLM_DATA_PREPARATION_MODULE_STORAGE_LOCATION = "3_llm_finetune_data/llm_ft_data.jsonl"

structlogger = structlog.get_logger()


@dataclass
class LLMDataExample:
    prompt: str
    output: List[PromptCommand]
    original_test_name: str
    original_user_utterance: str
    rephrased_user_utterance: str

    def as_dict(self) -> Dict[str, Any]:
        return {
            "prompt": self.prompt,
            "output": commands_as_string(self.output),
            "original_test_name": self.original_test_name,
            "original_user_utterance": self.original_user_utterance,
            "rephrased_user_utterance": self.rephrased_user_utterance,
        }


def _create_data_point(
    prompt: str,
    step: ConversationStep,
    conversation: Conversation,
    rephrased_user_message: Optional[str] = None,
) -> LLMDataExample:
    return LLMDataExample(
        prompt,
        step.llm_commands,
        conversation.get_full_name(),
        step.original_test_step.text,
        rephrased_user_message,
    )


async def _convert_conversation_into_llm_data(
    conversation: Conversation, agent: Agent
) -> List[LLMDataExample]:
    data = []

    # construct new conversations from the rephrasings
    new_conversations = _construct_new_conversations(conversation)

    original_user_steps = [
        step for step in conversation.iterate_over_annotated_user_steps()
    ]

    for i, step in enumerate(original_user_steps):
        # create data point for the original e2e test case
        data.append(_create_data_point(step.llm_prompt, step, conversation))

    test_case_name = conversation.name

    # create data points using the rephrasings, e.g. 'new_conversations'
    for rephrased_user_steps in new_conversations:
        sender_id = generate_sender_id(test_case_name)
        # create a new tracker to be able to simulate the conversation from start
        await agent.tracker_store.save(DialogueStateTracker(sender_id, slots=[]))
        # simulate the conversation to get the prompts
        for i, step in enumerate(original_user_steps):
            rephrased_user_message = rephrased_user_steps[i]
            user_message = UserMessage(rephrased_user_message, sender_id=sender_id)

            expected_commands = "\n".join(
                [command.to_dsl() for command in step.llm_commands]
            )
            fake_invoke_function = make_mock_invoke_llm(expected_commands)

            with (
                set_record_commands_and_prompts(),
                patch_invoke_llm_in_generators(fake_invoke_function),
            ):
                await agent.handle_message(user_message)

            rephrased_tracker = await agent.tracker_store.retrieve(sender_id)
            if rephrased_tracker is None:
                # if tracker doesn't exist, we can't create a data point
                continue

            latest_message = rephrased_tracker.latest_message
            if latest_message is None:
                # if there is no latest message, we don't create a data point
                continue

            # tell the type checker what we expect to find under "prompts"
            prompts = cast(
                Optional[List[Dict[str, Any]]], latest_message.parse_data.get(PROMPTS)
            )

            if prompts:
                # as we only use single step or compact command generator,
                # there is always exactly one prompt
                prompt = prompts[0]
                user_prompt: Optional[str] = prompt.get(KEY_USER_PROMPT)
                data.append(
                    _create_data_point(
                        user_prompt, step, conversation, rephrased_user_message
                    )
                )

    return data


def _construct_new_conversations(conversation: Conversation) -> List[List[str]]:
    """Construct new conversations from the rephrasings.

    In general, we will combine the passing rephrasings at the same index position to
    construct a new conversation. If for one particular user turn no other passing
    rephrasing exists, we reset the index and take the first passing rephrasing again.

    Args:
        conversation: The conversation.

    Returns:
        A list of new conversations (only rephrased user turns).
    """
    max_passed_rephrasings = max(
        [
            len(step.passed_rephrasings)
            for step in conversation.iterate_over_annotated_user_steps()
        ]
    )

    new_conversations = []
    for i in range(0, max_passed_rephrasings):
        current_conversation = []
        for step in conversation.iterate_over_annotated_user_steps():
            # take the orginial user message in case no passing rephrasings exist
            if not step.passed_rephrasings and step.original_test_step.text:
                structlogger.debug(
                    "llm_fine_tuning.llm_data_preparation_module."
                    "construct_new_conversations.no_passed_rephrasings",
                    conversation=conversation.get_full_name(),
                    step=step.original_test_step.text,
                    message="Take original user message instead of rephrasing.",
                )
                current_conversation.append(step.original_test_step.text)
                continue

            # some user steps might have fewer rephrasings than others
            # loop over the rephrasings
            index = i % len(step.passed_rephrasings)
            current_conversation.append(step.passed_rephrasings[index])
        if current_conversation:
            new_conversations.append(current_conversation)

    structlogger.debug(
        "llm_fine_tuning.llm_data_preparation_module.construct_new_conversations",
        conversation=conversation.get_full_name(),
        new_conversations=new_conversations,
    )

    return new_conversations


async def convert_to_fine_tuning_data(
    conversations: List[Conversation],
    storage_context: StorageContext,
    agent: Agent,
) -> List[LLMDataExample]:
    llm_data = []

    for i in tqdm(range(len(conversations))):
        conversation_llm_data = await _convert_conversation_into_llm_data(
            conversations[i], agent
        )
        llm_data.extend(conversation_llm_data)

    storage_context.write_llm_data(
        llm_data, LLM_DATA_PREPARATION_MODULE_STORAGE_LOCATION
    )

    return llm_data



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator

from rasa.core.constants import MCP_SERVERS_KEY
from rasa.shared.agents.auth.utils import validate_secrets_in_params
from rasa.shared.core.constants import (
    GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE,
    GLOBAL_SILENCE_TIMEOUT_KEY,
)
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import (
    EndpointConfig,
    read_endpoint_config,
    read_property_config_from_endpoints_file,
)


@dataclasses.dataclass
class InteractionHandlingConfig:
    """Configuration for interaction handling."""

    global_silence_timeout: Union[float, int] = GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE

    def __post_init__(self) -> None:
        # Validate the type of `global_silence_timeout`.
        if isinstance(self.global_silence_timeout, str):
            try:
                self.global_silence_timeout = float(self.global_silence_timeout)
            except ValueError:
                raise RasaException(
                    f"Type for {GLOBAL_SILENCE_TIMEOUT_KEY} is wrong, expected number. "
                    f"Got: '{self.global_silence_timeout}'. "
                )

        if not isinstance(self.global_silence_timeout, (float, int)):
            raise RasaException(
                f"Type for {GLOBAL_SILENCE_TIMEOUT_KEY} is wrong, expected number. "
                f"Got: '{type(self.global_silence_timeout)}'. "
            )

        if self.global_silence_timeout <= 0:
            raise RasaException(
                f"Value for {GLOBAL_SILENCE_TIMEOUT_KEY} must be a positive number. "
                f"Got: '{self.global_silence_timeout}'. "
            )

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]) -> InteractionHandlingConfig:
        """Create a InteractionHandlingConfig instance from a dictionary."""
        return cls(
            global_silence_timeout=data.get(
                GLOBAL_SILENCE_TIMEOUT_KEY, GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
            )
            if data is not None
            else GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
        )


class MCPServerConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str = Field(..., description="The name of the MCP server.")
    url: str = Field(..., description="The URL of the MCP server.")
    type: str = Field(..., description="The type of the MCP server.")
    additional_params: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Additional parameters for the MCP server."
    )

    @model_validator(mode="after")
    def validate_type(self) -> MCPServerConfig:
        # validate that type is "http"
        if self.type not in ["http", "https"]:
            raise ValueError(f"Invalid MCP server type: {self.type}")
        # validate that name and url are not empty
        if not self.name or not self.url:
            raise ValueError("Name and URL cannot be empty")
        # validate secrets in additional_params
        if self.additional_params:
            validate_secrets_in_params(
                self.additional_params, f"MCP server - '{self.name}'"
            )
        return self

    @model_validator(mode="before")
    def collect_additional_params(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        base_fields = {"name", "url", "type"}
        extras = {k: v for k, v in values.items() if k not in base_fields}
        if extras:
            values["additional_params"] = extras
            # remove them from top level so Pydantic doesn’t complain
            for k in extras:
                values.pop(k)
        return values


class AvailableEndpoints:
    """Collection of configured endpoints."""

    @classmethod
    def read_endpoints(cls, endpoint_file: Path) -> AvailableEndpoints:
        """Read the different endpoints from a yaml file."""
        nlg = read_endpoint_config(endpoint_file, endpoint_type="nlg")
        nlu = read_endpoint_config(endpoint_file, endpoint_type="nlu")
        action = read_endpoint_config(endpoint_file, endpoint_type="action_endpoint")
        model = read_endpoint_config(endpoint_file, endpoint_type="models")
        tracker_store = read_endpoint_config(
            endpoint_file, endpoint_type="tracker_store"
        )
        lock_store = read_endpoint_config(endpoint_file, endpoint_type="lock_store")
        event_broker = read_endpoint_config(endpoint_file, endpoint_type="event_broker")
        vector_store = read_endpoint_config(endpoint_file, endpoint_type="vector_store")
        raw_mcp_servers = read_property_config_from_endpoints_file(
            endpoint_file, property_name=MCP_SERVERS_KEY
        )
        mcp_servers = (
            [MCPServerConfig(**server) for server in raw_mcp_servers]
            if raw_mcp_servers
            else None
        )
        model_groups = read_property_config_from_endpoints_file(
            endpoint_file, property_name="model_groups"
        )
        privacy = read_property_config_from_endpoints_file(
            endpoint_file, property_name="privacy"
        )

        interaction_handling = InteractionHandlingConfig.from_dict(
            read_property_config_from_endpoints_file(
                endpoint_file, property_name="interaction_handling"
            )
        )

        return cls(
            endpoint_file,
            nlg,
            nlu,
            action,
            model,
            tracker_store,
            lock_store,
            event_broker,
            vector_store,
            mcp_servers,
            model_groups,
            privacy,
            interaction_handling,
        )

    def __init__(
        self,
        config_file_path: Optional[Path] = None,
        nlg: Optional[EndpointConfig] = None,
        nlu: Optional[EndpointConfig] = None,
        action: Optional[EndpointConfig] = None,
        model: Optional[EndpointConfig] = None,
        tracker_store: Optional[EndpointConfig] = None,
        lock_store: Optional[EndpointConfig] = None,
        event_broker: Optional[EndpointConfig] = None,
        vector_store: Optional[EndpointConfig] = None,
        mcp_servers: Optional[List[MCPServerConfig]] = None,
        model_groups: Optional[List[Dict[str, Any]]] = None,
        privacy: Optional[Dict[str, Any]] = None,
        interaction_handling: InteractionHandlingConfig = InteractionHandlingConfig(
            global_silence_timeout=GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
        ),
    ) -> None:
        """Create an `AvailableEndpoints` object."""
        self.config_file_path = config_file_path
        self.model = model
        self.action = action
        self.nlu = nlu
        self.nlg = nlg
        self.tracker_store = tracker_store
        self.lock_store = lock_store
        self.event_broker = event_broker
        self.vector_store = vector_store
        self.mcp_servers = mcp_servers
        self.model_groups = model_groups
        self.privacy = privacy
        self.interaction_handling = interaction_handling



from typing import NewType

# a common intermediate audio byte format that acts as a common data format,
# to prevent quadratic complexity between formats of channels, asr engines,
# and tts engines
# currently corresponds to raw wave, 8khz, 8bit, mono channel, mulaw encoding
RasaAudioBytes = NewType("RasaAudioBytes", bytes)
HERTZ = 8000
